// functions for error analysis and parameter adjustment
// SpinWorks 2.3


#include <stdio.h>
#include <stdlib.h>
#include "nummrit.h"
#include "maxvalue.h"
#include "matutil.h"
#include "trnout.h"
#include "sort_util.h"
#include "adjust.h"
#include "input_sim.h"
#include <math.h>


// /grpcom/
extern double	*cshift, **cup, **symcup,
				**dip, **symdip;
extern int		*grpiso, *nucg, *pairg,
				*grpspn, fmaxg[MAXGRP+1][MAXGRP+1], *agpb;
extern short	anydip, symspn;
extern int		nm2, km2, ne2, nm32, sym2, wgt2,
				*amaxg, *aming, *ag,
				**fg;

extern char		grpnam[MXGNAM + 1][MAXGRP + 1];

// /asncom/
extern double	*asnfrq, *asnwgt, *asndif, *asnclc, asnint;
extern int		asncnt, *asnlin;

// /flgcom/
extern short	autasn, autdel, optmiz, prtptl,
				prtkey, prtdif, prtvcm, pracnd;

// /optcom/
extern double	rmscon, rmsdiv, rmslim, damp, asnlim,
				ignlim, itrfac, linwid, asnrms;
extern int		itrlim, itrmin, asnmod;
extern short	anyasn, wgtrms;

// /spncom/
extern int		ngrp, niso, nvar;

// /obscom/
extern int		obscnt, *obslin;
extern double	*obsfrq, *obsint;

// /prmcom/
extern int		prmsiz, prmnum, prmi, 
				prmj, prmd, lowvar;
extern double	rngdef, rngnit, rngmin;
extern double	*asnhi, *asnlow,
				*rhics,	*rlocs, rngfra,
				**rhiccj, **rloccj,
				**rhiscj, **rloscj,
				**rhiccd, **rloccd,
				**rhiscd, **rloscd;
extern int		*asnmin, *asnnum,
				**prmmap, *ivcs,
				**ivcup, **ivdip;

//  iteration parameters
double rms, rms_old;
int itrcnt, delcnt;
short iting;

// optimized couplings and shifts, for loading into spin system editor
double j_coup_opt[MAX_SHIFTS][MAX_SHIFTS];
double d_coup_opt[MAX_SHIFTS][MAX_SHIFTS];
double shift_opt[MAX_SHIFTS];

//extern CString sim_out;

//#define DUMP_LOG

/////////////////////////////////////////////////////////////////////////////////

int adjust(const char *sim_out)
{
	// main adjustment control routine
	int error = FALSE;

	converge(sim_out);

	if(iting)
		error = newvar(sim_out);
	else
		 erranal(sim_out);

	return error;
}

////////////////////////////////////////////////////////////////////////////////

void converge(const char *sim_out)
{
	// convergence test routine
	double templim;
	short altered, assign;
	FILE *ofp;
	

	if(itrcnt == 0) delcnt = 0;	// can't delete any on first iteration

	trnrms(sim_out);	// get the rms value

	if(!iting)	// have terminated so print final iteration count
	{
		ofp = fopen(sim_out, "a");
		fprintf(ofp, "\n***  After final iteration:\n");
		fprintf(ofp, "***      current iteration count = \t%8d\n", itrcnt);
		fprintf(ofp, "***      transitions now assigned = \t%8d\n", asncnt);
		fprintf(ofp, "***      current RMS = \t\t\t%f\n", rms);
		fclose(ofp);
	}
	else		// still going
	{
		altered = FALSE;

		assign = autasn && (rms <= asnrms) 
			            && ((itrcnt % (asnmod+1)) == asnmod
						  || (!anyasn && (itrcnt > asnmod)));

		if(assign)		// autoassign the lines
		{
			templim = asnlim * rms;
			trnasn(sim_out, templim);
			setaut();
			altered = TRUE;
			anyasn = TRUE;
			itrmin = itrcnt + asnmod - 1;
			delcnt = 0;
		}
					
		else if(autdel)	// autodelete if requested, but only in we haven't
		{				// autoassigned this iteration
			templim = ignlim * rms;
			if(itrcnt == 0) templim = templim * 5.0;
			trndel(sim_out, templim, anyasn, &altered);
		}

		ofp = fopen(sim_out, "a");
		fprintf(ofp, "\n***   Transitions now assigned = %8d\n", asncnt);
		fclose(ofp);

		if((!altered && (itrcnt > itrlim))	// if we are over iteration limit and
			|| (itrcnt > itrlim + 3))	
		{									// no autoassign or autodelete has occured 
			iting = FALSE;					// then quit
			ofp = fopen(sim_out, "a");
			fprintf(ofp, "\n***  no convergence after = %3d iterations\n", itrcnt);
			fclose(ofp);
		}
		else								// we can keep iterating
		{
			if(altered)
			{
				//error = prmchk();
			}
			itrcnt++;						// bump the iteration counter
		}
	}
}

////////////////////////////////////////////////////////////////////////////////

int newvar(const char *sim_out)
{
//	This routine solves the least squares normal equations and
//	uses the solution to update the values of the shifts and
//	couplings. The revised parameters are then printed.
	int		error = FALSE;

	double	**normeq;
	double	*var, *soln;
	double	**cupnew, **cuperr, **dipnew, **diperr, 
			**symcer, **symder, *cserr;

	int		i, j, k, l, ivprm;
	short	vboth, vnone, npair;

	FILE *ofp;

#ifdef DUMP_LOG
	FILE *fp;
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** newvar(): local variables\n");
	fclose(fp);
#endif 
	normeq = matrix(1, nvar, 1, nvar);
	var = vector(1, nvar);
	soln = vector(1, nvar);
	cupnew = matrix(1, ngrp, 1, ngrp);
	cuperr = matrix(1, ngrp, 1, ngrp);
	dipnew = matrix(1, ngrp, 1, ngrp);
	diperr = matrix(1, ngrp, 1, ngrp);
	symcer = matrix(1, ngrp, 1, ngrp);
	symder = matrix(1, ngrp, 1, ngrp);
	cserr = vector(1, ngrp);

	dclear(nvar, var, 0.0);
	dclear(nvar, soln, 0.0);

//	ivprm > 0 : parameter can be optimized
//	iv... > 0 : parameter is independent

	for(i=1; i <= ngrp; i++)
	{
		ivprm = prmmap[prmnum][abs(ivcs[i])];
		if(ivprm > 0 && ivcs[i] > 0) var[ivcs[i]] = cshift[i];
		for(j = 1; j <= ngrp; j++)
		{
			ivprm = prmmap[prmnum][abs(ivcup[i][j])];
			if(ivprm > 0 && ivcup[i][j] > 0)
			{
				var[ivcup[i][j]] = symcup[i][j];
			}

			ivprm = prmmap[prmnum][abs(ivdip[i][j])];
			if(ivprm > 0 && ivdip[i][j] > 0)
			{
				var[ivdip[i][j]] = symdip[i][j];
			}
		}
	}


//	solve the normal equations and adjust the variable parameters

	setnrm(normeq, soln);
	error = regres(normeq, soln);

	if(error == ERROR)		// hope we never see the following!
	{
		ofp = fopen(sim_out, "a");
		fprintf(ofp, "   \n\n*** ERROR:  could not solve normal equations!\n");
		fprintf(ofp, "       Optimization is terminated\n");
		fprintf(ofp, "       Final parameters will NOT be printed\n\n");
		fclose(ofp);
		return error;		// bail out
	}

	for(j=1; j<= nvar; j++) 
	{
		var[j] = var[j] + soln[j] * damp;
	}


//	Transform variables to new shifts and symmetric couplings.
//	This applies to both independent and dependent variables.
//	We make sure the variables are within the user specified
//	ranges. Note that for couplings, we use the original
//	symmeterized ranges (could be a problem?).

	ofp = fopen(sim_out, "a");	/////
	for(i=1; i <= ngrp; i++)
	{
		k = prmmap[prmnum][abs(ivcs[i])];
		if(k > 0) 
		{
			cshift[i] = minmax(rlocs[i], var[k], rhics[i]);
		}

		k = prmmap[prmnum][abs(ivcup[i][i])];
		if(k > 0)
		{
			symcup[i][i] = minmax(rloscj[i][i], var[k], rhiscj[i][i]);
		}

		if(anydip)
		{
			k = prmmap[prmnum][abs(ivdip[i][i])];
			if(k > 0)
				symdip[i][i] = minmax(rloscd[i][i], var[k], rhiscd[i][i]);
		}

		for(j = i+1; j <= ngrp; j++)
		{
			k = prmmap[prmnum][abs(ivcup[i][j])];
			l = prmmap[prmnum][abs(ivcup[j][i])];
			vboth = (k > 0) && (l > 0);
			vnone = (k == 0) && (l == 0);
			npair = (pairg[i] != 2) || (pairg[j] != 2);

			if(vboth || vnone || npair)	// ordinary coupling
			{
				if(k > 0)
					symcup[i][j] = minmax(rloscj[i][j], var[k], rhiscj[i][j]);

				if(l > 0)
						symcup[j][i] = minmax(rloscj[j][i], var[l], rhiscj[j][i]);
			}
			else						// only half a symmetric pair is optimized
			{
				if(l > 0)
				{
					symcup[i][j] = minmax(rloscj[i][j], cup[i][j] - var[l],
						rhiscj[i][j]);
					symcup[j][i] = minmax(rloscj[j][i], var[l], rhiscj[j][i]);
				}
				else
				{
					symcup[i][j] = minmax(rloscj[i][j], var[k], rhiscj[i][j]);
					symcup[j][i] = minmax(rloscj[j][i], var[k] - cup[j][i], 
						rhiscj[j][i]);
				}
			}

			if(anydip)
			{
				k = prmmap[prmnum][abs(ivdip[i][j])];
				l = prmmap[prmnum][abs(ivdip[j][i])];
				vboth = (k > 0) && (l > 0);
				vnone = (k == 0) && (l == 0);
				npair = (pairg[i] != 2) || (pairg[j] != 2);

				if(vboth || vnone || npair)	// ordinary coupling
				{
					if(k > 0)
						symdip[i][j] = minmax(rloscd[i][j], var[k], rhiscd[i][j]);
					if(l > 0)
						symdip[j][i] = minmax(rloscd[j][i], var[l], rhiscd[j][i]);
				}	
				else						// only half a symmetric pair is optimized
				{
					if(l > 0)
					{
						symdip[i][j] = minmax(rloscd[i][j], dip[i][j] - var[l],
							rhiscd[i][j]);
						symdip[j][i] = minmax(rloscd[j][i], var[l], rhiscd[j][i]);
					}
					else
					{
						symdip[i][j] = minmax(rloscd[i][j], var[k], rhiscd[i][j]);
						symdip[j][i] = minmax(rloscd[j][i], var[k] - dip[j][i], 
							rhiscd[j][i]);
					}
				}
			}

		}	// 70
	}		// 60
	fclose(ofp);

	seterr(cserr, symcer, symder, soln);
	symfrm(ngrp, pairg, symcup, cupnew);
	symfrm(ngrp, pairg, symcer, cuperr);

	if(anydip)
	{
		symfrm(ngrp, pairg, symdip, dipnew);
		symfrm(ngrp, pairg, symder, diperr);
	}

	ofp = fopen(sim_out, "a");
	fprintf(ofp, "\n***  Iteration # %d, new parameters are:\n", itrcnt); 
	for(i = 1; i<=ngrp; i++)
	{
		if(ivcs[i] != 0)
	  	    fprintf(ofp, "   v[%d] = \tv[%s]      \t\t%10.3f change: \t%7.4f\n", 
				i, grpnam[i], cshift[i], cserr[i]);
		else
			fprintf(ofp, "   v[%d] = \tv[%s] \t\t%10.3f \n", 
				i, grpnam[i], cshift[i]);
	}
	for(i = 1; i <= ngrp; i++)
		for(j = 1; j <= ngrp; j++)
		{
			if((j > i) || (pairg[i] == 2 && pairg[j] == 2))
			{
				if(ivcup[i][j] != 0)
					fprintf(ofp, "   j[%d][%d] = \tj[%s][%s]\t\t%10.3f change: \t%7.4f\n", 
						i, j, grpnam[i], grpnam[j], cupnew[i][j], cuperr[i][j]);
				else
					fprintf(ofp, "   j[%d][%d] = \tj[%s][%s]\t\t%10.3f\n", 
						i, j, grpnam[i], grpnam[j], cupnew[i][j]);
			}
		}
	if(anydip)
	{
		for(i = 1; i <= ngrp; i++)
			for(j = 1; j <= ngrp; j++)
			{
				if(j > i || (j == i && nucg[i] > 1))
				{
					if(ivdip[i][j] != 0)
						fprintf(ofp, "   d[%d][%d] = \td[%s][%s]\t\t%10.3f change: \t%7.4f\n", 
							i, j, grpnam[i], grpnam[j], dipnew[i][j], diperr[i][j]);
					else
						fprintf(ofp, "   d[%d][%d] = \td[%s][%s]\t\t%10.3f\n", 
							i, j, grpnam[i], grpnam[j], dipnew[i][j]);
				}
			}
	}

	fprintf(ofp, "\n");
	fclose(ofp);

//  free memory
#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** newvar(): freeing local variables\n");
	fclose(fp);
#endif 
	free_matrix(normeq, 1, nvar, 1, nvar);
	free_vector(var, 1, nvar);
	free_vector(soln, 1, nvar);
	free_matrix(cuperr, 1, ngrp, 1, ngrp);
	free_matrix(dipnew, 1, ngrp, 1, ngrp);
	free_matrix(diperr, 1, ngrp, 1, ngrp);
	free_matrix(symcer, 1, ngrp, 1, ngrp);
	free_matrix(symder, 1, ngrp, 1, ngrp);
	free_vector(cserr, 1, ngrp);

	return error;
}

////////////////////////////////////////////////////////////////////////////////

void erranal(const char *sim_out)
{
	double sigma;
	int i, j;
	int error;

	double	**normeq, *tmprhs, **vcm, **corr, *stand_err, 
			*cserr, **cuperr, **symcer, **diperr, **symder;

//  allocate array space
#ifdef DUMP_LOG
	FILE *fp;
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** erranal(): local variables\n");
	fclose(fp);
#endif 
	normeq = matrix(1, nvar, 1, nvar);
	tmprhs = vector(1, nvar);
	vcm = matrix(1, nvar, 1, nvar);
	corr = matrix(1, nvar, 1, nvar);
	stand_err = vector(1, nvar);

	cserr = vector(1, ngrp);
	cuperr = matrix(1, ngrp, 1, ngrp);
	symcer = matrix(1, ngrp, 1, ngrp);
	diperr = matrix(1, ngrp, 1, ngrp);
	symder = matrix(1, ngrp, 1, ngrp);

	sigma = rms * rms;

	// initialize matrices
	dclear(ngrp, cserr, 0.0);
	matclear(cuperr, 1, ngrp, 1, ngrp, 0.0);
	matclear(diperr, 1, ngrp, 1, ngrp, 0.0);

	dclear(nvar, stand_err, 0.0);
	matclear(vcm, 1,nvar, 1, nvar, 0.0);
	matclear(corr, 1, nvar, 1, nvar, 0.0);

	setnrm(normeq, tmprhs);
	error = invert(nvar, normeq, vcm);

	for(i = 1; i <= nvar; i++)
		for(j = 1; j <= nvar; j++)
			vcm[i][j] = sigma * vcm[i][j];

	//  Get the standard deviations of the variable parameters and the
	//  correlation coefficients between them.

	for(i=1; i <=nvar; i++)
	{
		stand_err[i] = sqrt(fabs(vcm[i][i]));
		for(j = 1; j <= i; j++)
			corr[i][j] = vcm[i][j]/(stand_err[i]*stand_err[j]);
	}

	//  Transform the standard deviation in variable parameters to
	//  the standard deviation in spectral parameters

	seterr(cserr, symcer, symder, stand_err);

	//  Normalize  symmetric couplings and errors for systems with symmetry

	symfrm(ngrp, pairg, symcup, cup);
	symfrm(ngrp, pairg, symcer, cuperr);
	symfrm(ngrp, pairg, symdip, dip);
	symfrm(ngrp, pairg, symder, diperr);

	setlwd();

	//  Ensure that errors are all positive as there can be problems with
	//  those of symmetric pairs.

	for(i = 1; i <= ngrp; i++)
		for(j = 1; j <= ngrp; j++)
		{
			cuperr[i][j] = fabs(cuperr[i][j]);
			diperr[i][j] = fabs(diperr[i][j]);
		}


	//  Print Final Parameters, and save optimized parameters for future use.

	FILE *ofp = fopen(sim_out, "a");
	fprintf(ofp, "\n***  Final parameters after %d iterations are:\n", itrcnt);
	
	for(i = 1; i<=ngrp; i++)	// print shifts
	{
		if(ivcs[i] != 0)
	  		fprintf(ofp, "   v[%d] = \tv[%s]      \t\t%10.3f Hz. +/- \t%7.4f Hz.\n", 
				i, grpnam[i], cshift[i], cserr[i]);
		else
			fprintf(ofp, "   v[%d] = \tv[%s] \t\t%10.3f Hz.\n", 
				i, grpnam[i], cshift[i]);

			shift_opt[i-1] = cshift[i];
	}

	for(i = 1; i <= ngrp; i++)	// print scalar couplings
		for(j = 1; j <= ngrp; j++)
		{
			if((j > i) || (pairg[i] == 2 && pairg[j] == 2))
			{
				if(ivcup[i][j] != 0)
					fprintf(ofp, "   j[%d][%d] = \tj[%s][%s]\t\t%10.3f Hz. +/- \t%7.4f Hz.\n", 
						i, j, grpnam[i], grpnam[j], cup[i][j], cuperr[i][j]);
				else
					fprintf(ofp, "   j[%d][%d] = \tj[%s][%s]\t\t%10.3f Hz.\n", 
						i, j, grpnam[i], grpnam[j], cup[i][j]);

				j_coup_opt[i-1][j-1] = cup[i][j];
			}
		}

	for(i = 1; i <= ngrp; i++)	// print dipolar couplings
		for(j = 1; j <= ngrp; j++)
			if(fabs(dip[i][j]) > 0.0)
			{
				if(ivdip[i][j] !=0)
					fprintf(ofp, "   d[%d][%d] = \tj[%s][%s]\t\t%10.3f Hz. +/- \t%7.4f Hz.\n", 
						i, j, grpnam[i], grpnam[j], dip[i][j], diperr[i][j]);
				else
					fprintf(ofp, "   d[%d][%d] = \tj[%s][%s]\t\t%10.3f Hz.\n", 
						i, j, grpnam[i], grpnam[j], dip[i][j]);

				d_coup_opt[i-1][j-1] = dip[i][j];
			}

	fprintf(ofp, "\n***  Parameter correlation coefficients:\n");

	for(i = 1; i <=nvar; i++)
	{
		fprintf(ofp, "  %2d", i);
		for(j = 1; j <= i; j++)
		{
			fprintf(ofp, " %6.3f", corr[i][j]);
		}
		fprintf(ofp, "\n");
	}

	fprintf(ofp, "\n");
	fclose(ofp);

	//  free up memory
	free_matrix(normeq, 1, nvar, 1, nvar);
	free_vector(tmprhs, 1, nvar);
	free_matrix(vcm, 1, nvar, 1, nvar);
	free_matrix(corr, 1, nvar, 1, nvar);
	free_vector(stand_err, 1, nvar);

	free_vector(cserr, 1, ngrp);
	free_matrix(cuperr, 1, ngrp, 1, ngrp);
	free_matrix(symcer, 1, ngrp, 1, ngrp);
	free_matrix(diperr, 1, ngrp, 1, ngrp);
	free_matrix(symder, 1, ngrp, 1, ngrp);

}

////////////////////////////////////////////////////////////////////////////////

void trnrms(const char *sim_out)
{

	const int frcitr = 5, chkitr = 8;
	const double frcchk= 7.5e-4, fmtlim = 1000.0;

	double rmsfra, absdif, bigdif = 0.0, sumsqr = 0.0, weight = 1.0,
		   totwgt = 0.0, frcavg;

	static double frctot;

	int i, unbias;	
	FILE *ofp = fopen(sim_out, "a");

	for(i=1; i <= asncnt; i++)
	{
		asndif[i] = asnfrq[i] - asnclc[i];
		absdif = fabs(asndif[i]);
		if(absdif > bigdif) bigdif = absdif;
		if(wgtrms == TRUE) weight = fabs(asnwgt[i]);
		sumsqr = sumsqr + absdif * absdif * weight;
		totwgt = totwgt + weight;
	}

	if(itrcnt != 0) rms_old = rms;
	unbias = asncnt - nvar;
	if(unbias <= 0) unbias = asncnt;
	rms = sqrt(sumsqr/(double)unbias);

//	prtdif = TRUE;	////////////
	if(prtdif || itrcnt == 0)
	{
		
		fprintf(ofp, "\n line\t\t obs. freq.\t calc. freq\t diff.\n\n");
		if(bigdif < fmtlim)
		{
			for(i=1; i <= asncnt; i++)
				fprintf(ofp, " %8d\t%10.3f\t%10.3f\t%9.3f\n",
					asnlin[i], asnfrq[i], asnclc[i], asndif[i]);   
		}
		else
		{
			for(i=1; i <= asncnt; i++)
				fprintf(ofp, " %8d\t%10.3f\t%10.3f\t%10.2e\n",
					asnlin[i], asnfrq[i], asnclc[i], asndif[i]);   
		}
		
	}

	if(bigdif < fmtlim && rms < fmtlim && asnint < fmtlim)
	{
	fprintf(ofp, "   Largest absolute difference  = %9.3f\n", bigdif);
	fprintf(ofp, "   Total assigned intensity     = %9.3f\n", asnint);
	fprintf(ofp, "   RMS deviation of transitions = %9.5f\n", rms);
	}
	else
	{
	fprintf(ofp, "   Largest absolute difference  = %12.3e\n", bigdif);
	fprintf(ofp, "   Total assigned intensity     = %12.3e\n", asnint);
	fprintf(ofp, "   RMS deviation of transitions = %12.3e\n", rms);
	}

	

	if(itrcnt == 0)
		frctot = 0.0;
	else
	{
		//  The fractional change is negative if converging and positive
		//  if diverging.  rmscon and rmsdiv are however both positive.

		rmsfra = (rms - rms_old) / rms_old;
		if(fabs(rmsfra) < fmtlim)
			fprintf(ofp, "   Fractional change of RMS     = %9.5f\n", rmsfra);
		else
			fprintf(ofp, "   Fractional change of RMS     = %12.3e\n", rmsfra);

		if(itrcnt > frcitr)  // average a few iterations
		{
			frctot = frctot + fabs(rmsfra);
			frcavg = frctot / (double)(itrcnt-frcitr);
		}

		if(itrcnt > itrmin)
		{
			if(rmsfra > rmsdiv)
			{
				iting = FALSE;
				if(rms  <= 2.0 * rmslim)
				{
					fprintf(ofp, "\n*** Convergence achieved (fractional increase) ***\n");
				}
				else
				{
					fprintf(ofp, "\n*** Divergence! ***\n");
					fprintf(ofp, "   RMS deviation increased from %11.4f to %11.4f\n",
						rms_old, rms);
					fprintf(ofp, "   during iteration %4d\n", itrcnt);
				}
			}

			else if(rms < rmslim)
			{
				iting = FALSE;
				fprintf(ofp, "\n*** Convergence achieved (RMS Limit)\n");
			}

			else if((!autasn && fabs(rmsfra) < rmscon) ||
				    (autasn && rms <= (2.0 * rmslim)
					        && fabs(rmsfra) < rmscon))
			{
				iting = FALSE;
				fprintf(ofp, "\n*** Convergence achieved (Fractional decrease)\n");
			}

			else if(itrcnt >= chkitr && frcavg <= frcchk)
			{
				iting = FALSE;
				fprintf(ofp, "\n*** Convergence achieved (Parameters not changing)\n");
				fprintf(ofp, "      This could be due to insufficient assigned transitions in a second-order spectrum\n");
			}
		}
	}	// itrcnt != 0

	fclose(ofp);
}

////////////////////////////////////////////////////////////////////////////////

void trnasn(const char *sim_out, double asnwid)
{
//     This routine performs the assignment of observed transitions
//     to calculated transitions. This is done separately for each
//     isotopic species. Calculated transitions within asnwid Hertz
//     of an observed transitions are assigned to that observed
//     transition. Global asncnt returns the total number of transitions
//     assigned.

	double	*calc_freq, *calc_ampl; 
	int		*calc_line, *calc_key;

	int		i, clccnt, ixclc, ixobs, ixgrp;
	double	isoamp, freq, intens, cuthif, cutlof, cfreq;
	short	asnabl;
	FILE *ofp = fopen(sim_out, "a");

#ifdef DUMP_LOG
	FILE *fp;
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** trnasn(): local variables\n");
	fclose(fp);
#endif 
	calc_freq = vector(1, MAXCLC);
	calc_ampl = vector(1, MAXCLC);
	calc_line = ivector(1, MAXCLC);
	calc_key = ivector(1, MAXCLC);

	dclear(MAXCLC, calc_freq, 0.0);
	dclear(MAXCLC, calc_ampl, 0.0);
	iclear(MAXCLC, calc_line, 0);
	iclear(MAXCLC, calc_key, 0);

//	fprintf(ofp, "\n***  trnasn(): entering routine\n");
	asncnt = 0;

	for(i = 1; i <= niso; i++)
	{
		tran_get(i, MAXCLC, &clccnt, &isoamp, calc_freq, calc_ampl, calc_line, calc_key);
		shell_sort_trans(MAXCLC, clccnt, calc_freq, calc_ampl, calc_line, calc_key);
//		fprintf(ofp, "***  trnasn(): clccnt = %d     obscnt = %d    asncnt = %d   asnwid = %9.3f\n\n", 
//			clccnt, obscnt, asncnt, asnwid);
		ixobs = 1;
		ixclc = 1;

		while(ixobs <= obscnt && ixclc <= clccnt && asncnt < MAXASN)
		{
			intens = obsint[ixobs];
			freq = obsfrq[ixobs];
			asnabl = FALSE;

			for(ixgrp = 1; ixgrp <= ngrp; ixgrp ++)
			{
				if((freq >= asnlow[ixgrp] && freq <= asnhi[ixgrp])
					&& ivcs[ixgrp] > 0)
				{
					asnabl = TRUE;
		//			fprintf(ofp, "***  trnasn(): asnabl!  low = %9.3f  freq = %9.3f  hi = %9.3f\n",
		//				asnlow[ixgrp], freq, asnhi[ixgrp]);
					break;
				}
			}

			if(asnabl == TRUE)
			{
				cuthif = freq + asnwid;
				if(ixobs < obscnt)
				{
					cuthif = fmin(cuthif, obsfrq[ixobs+1]);
				}

				cutlof = freq - asnwid;
				if(ixobs > 1)
				{
					cutlof = fmax(cutlof, obsfrq[ixobs-1]);
				}

				while(ixclc <= clccnt && asncnt < MAXASN)
				{
					cfreq = calc_freq[ixclc];

					if(cfreq > cuthif) 
					{
						break;
					}

					if(cfreq >= cutlof)
					{
						asncnt++;
						asnfrq[asncnt] = freq;
						asnwgt[asncnt] = intens;
						asnlin[asncnt] = calc_line[ixclc];
						asnclc[asncnt] = calc_freq[ixclc];
						asndif[asncnt] = freq - calc_freq[ixclc];
					}
					ixclc++;
				}		// 310
			}			// endif
			ixobs++;
		}				// 210
	}					// 100

	stasnl(MAXASN, asncnt, asnlin, asnfrq, asnwgt, asnclc, asndif);

	//  free up space
	free_vector(calc_freq, 1, MAXCLC);
	free_vector(calc_ampl, 1, MAXCLC);
	free_ivector(calc_line, 1, MAXCLC);
	free_ivector(calc_key, 1, MAXCLC);

	fprintf(ofp, "\n***   Automatic assignment has occurred  ***\n");
	fclose(ofp);
}

////////////////////////////////////////////////////////////////////////////////

void trndel(const char *sim_out, double cutoff, short anyasn, short *altered)
{
	const double maxdif = 100.0;
	const double freqlim = 1000000.0;

	double absdif;
	int cnttmp, limit, i;
	FILE *ofp = fopen(sim_out, "a");

	limit = DELPCT * (asncnt - nvar)/100;
	cnttmp = asncnt;

//	fprintf(ofp, "\n***  trndel(): limit = %d\n", limit);
	fprintf(ofp, "\n");

	if(delcnt < limit)
	{
		asncnt = 0;
		for(i = 1; i <= cnttmp; i++)
		{
			absdif = fabs(asndif[i]);
//			fprintf(ofp, "***  trndel(): asndif[%d] = %f\n", i, asndif[i]);
			if((absdif >= cutoff || absdif >= maxdif) && delcnt < limit)
			{
//				fprintf(ofp, "           deleting!\n");
				delcnt++;
				if(!anyasn)
				{
					if(fabs(asnfrq[i]) < freqlim && 
					   fabs(asndif[i]) < freqlim)
					{
						fprintf(ofp, 
					     "***  Ignored trans # %6d with freq. = %10.3f Hz; difference = %10.3f Hz\n",
							asnlin[i], asnfrq[i], asndif[i]);
					}
					else
					{
						fprintf(ofp, 
					     "***  Ignored trans # %6d with freq. = %15.1f Hz; difference = %15.1f Hz\n",
							asnlin[i], asnfrq[i], asndif[i]);
					}
				}
			}
			else
			{
				asncnt++;
				asnfrq[asncnt] = asnfrq[i];
				asnwgt[asncnt] = asnwgt[i];
				asnlin[asncnt] = asnlin[i];
				asndif[asncnt] = asndif[i];
				asnclc[asncnt] = asnclc[i];
			}
		}	// 200
	}
	*altered = (asncnt != cnttmp);
	fprintf(ofp, "\n***  Ignored %5d transitions\n", cnttmp - asncnt);
	fclose(ofp);
}

////////////////////////////////////////////////////////////////////////////////

void setaut()
{
//	This routine adjusts the parameters that control automatic
//	transition assignment and deletion. This is an attempt to
//	avoid excess iterations.

	ignlim = itrfac * ignlim;
	asnlim = itrfac * asnlim;
	if(ignlim < 1.6) ignlim = 1.6;
	if(asnlim > ignlim) asnlim = 0.90 * ignlim;
}

//////////////////////////////////////////////////////////////////////////////////

void setnrm(double **normeq, double *rhs)
{
//	This routine reads the partial differentials of observed
//	transitions w.r.t. parameters from unit IDIF and creates
//	the least squares normal equations. If a transition exists
//	on disk but not in the list of assigned transitions, it is
//	assumed that the transition has been ignored.

	double	*pd, weight, difval;
	int		*isp, i, j, ixasn, itn;
	FILE	*scratch4;

#ifdef DUMP_LOG
	FILE	*fp;
#endif

#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** setnrm(): local variables\n");
	fclose(fp);
#endif 
	pd = vector(1, MAXVAR);			//  allocate space
	isp = ivector(1, MAXVAR);		//  goes all the way to maxvar because
									//  the IDIF file is written that way!
	iclear(nvar, isp, 1);
	matclear(normeq, 1, nvar, 1, nvar, 0.0);
	dclear(nvar, rhs, 0.0);

	for(i = 1; i <= ngrp; i++)
	{
		if(ivcs[i] < 0) isp[-ivcs[i]] = isp[-ivcs[i]] + 1;
		for(j = 1; j <= ngrp; j++)
		{
			if(ivcup[i][j] < 0)
				isp[-ivcup[i][j]] = isp[-ivcup[i][j]] +1;
			if(ivdip[i][j] < 0)
				isp[-ivdip[i][j]] = isp[-ivdip[i][j]] +1;
		}
	}

	scratch4 = fopen("scratch.4", "rb");	// IDIF
	weight = 1.0;

	while(!feof(scratch4) && !ferror(scratch4))
	{
		fread(&itn, sizeof(int), 1, scratch4);		// IDIF
		if(feof(scratch4) || ferror(scratch4)) break;
		
		for(i = 1; i <= MAXVAR; i++)				// why all the way to MAXVAR?			
			fread(&pd[i], sizeof(double), 1, scratch4);

		ixasn = binint(MAXASN, asncnt, asnlin, itn);

		if(ixasn > 0 && ixasn <= asncnt)
		{
			if(wgtrms == TRUE) weight = fabs(asnwgt[ixasn]);
			difval = asndif[ixasn] * weight;
			for(j = 1; j <= nvar; j++)
			{
				rhs[j] = rhs[j] + difval * pd[j] * (double)isp[j];
				for(i = 1; i <= nvar; i++)
				{
					normeq[i][j] = normeq[i][j]  
						         + pd[i] * pd[j] * (double)(isp[i] * isp[j]);
				}	// 80
			}	// 70
		}
	}

	free_vector(pd, 1, MAXVAR);						//  free space and clean up
	free_ivector(isp, 1, MAXVAR);
	fclose(scratch4);								// IDIF
}

//////////////////////////////////////////////////////////////////////////////////

void seterr(double *cserr, double **cuperr, double **diperr, double *errvec)
{
	int i, j, iv;

	dclear(ngrp, cserr, 0.0);
	matclear(cuperr, 1, ngrp, 1, ngrp, 0.0);
	matclear(diperr, 1, ngrp, 1, ngrp, 0.0);
	for(i = 1; i <= ngrp; i++)
	{
		iv = abs(ivcs[i]);
		if(prmmap[prmnum][iv] != 0) cserr[i] = errvec[iv];

		iv = abs(ivcup[i][i]);
		if(prmmap[prmnum][iv] != 0) cuperr[i][i] = errvec[iv];

		iv = abs(ivdip[i][i]);
		if(prmmap[prmnum][iv] != 0) diperr[i][i] = errvec[iv];

		for(j = i+1; j <= ngrp; j++)
		{
			iv = abs(ivcup[i][j]);
			if(prmmap[prmnum][iv] != 0) cuperr[i][j] = errvec[iv];
			iv = abs(ivcup[j][i]);
			if(prmmap[prmnum][iv] != 0) cuperr[j][i] = errvec[iv];

			iv = abs(ivdip[i][j]);
			if(prmmap[prmnum][iv] != 0) diperr[i][j] = errvec[iv];
			iv = abs(ivdip[j][i]);
			if(prmmap[prmnum][iv] != 0) diperr[j][i] = errvec[iv];
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////

int regres(double **mat, double *rhs)
{
//	Solve the linear set of equations for matrix MAT and righthand
//	side RHS. ERROR is true if the matrix is singular. MAT is
//	not modified (ie not returned as its LU decomposition)!

	double	**wrkmat, *scale, xchng;
	int		*pindx;
	int		error = FALSE;

#ifdef DUMP_LOG
	FILE	*fp;
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** regres(): local variables\n");
	fclose(fp);
#endif 
	wrkmat = matrix(1, nvar, 1, nvar);
	scale = vector(1, nvar);
	pindx = ivector(1, nvar);

	//  make a copy of matrix
	int i,j;
	for(i = 1; i <= nvar; i++)
		for(j = 1; j <= nvar; j++)
			wrkmat[i][j] = mat[i][j];

	error = ludcmp(nvar, wrkmat, pindx, scale, &xchng);

	if(error == FALSE)
		ludbak(nvar, wrkmat, pindx, rhs);

	free_matrix(wrkmat, 1, nvar, 1, nvar);
	free_vector(scale, 1, nvar);
	free_ivector(pindx, 1, nvar);
	return error;
}

//////////////////////////////////////////////////////////////////////////////////

void group_rms(const char *sim_out)
{
	const double fmtlim = 1000.0;

	FILE *outfile;

	int		i, unbias, grplin, grpcnt, grptot = 0;

	double	sumsqr = 0, weight = 1.0, totwgt = 0.0, 
			grpfrq, grpdif, grpwgt, rms;

	shell_sort_asn(MAXASN, asncnt, asnlin, asnfrq, asnwgt, asnclc, asndif);

	i = 1;

	while(i <= asncnt)
	{
		group_assign(linwid, &i, &grpfrq, &grplin, &grpdif, &grpwgt, &grpcnt);
		grptot++;
		if(wgtrms) weight = fabs(grpwgt);
		sumsqr = sumsqr + grpdif * grpdif * weight;
		totwgt = totwgt + weight;
	}

		
	unbias = grptot - nvar;
	if(unbias <= 0) unbias = grptot;
	rms = sqrt(sumsqr / unbias);

	outfile = fopen(sim_out, "a");

	if(rms < fmtlim)
		fprintf(outfile, "\n  RMS deviation of %5d peaks grouped within %9.3f Hz = %9.5f Hz\n\n",
		grptot, linwid, rms);
	else
		fprintf(outfile, "\n  RMS deviation of %5d peaks grouped within %9.3f Hz = %12.3f Hz\n\n",
		grptot, linwid, rms);

	fclose(outfile);

}

////////////////////////////////////////////////////////////////////////////////

void group_assign(double grpwid, int *curix, double *grpfrq, int *grplin,
				  double *grpdif, double *grpwgt, int *grpcnt)
{
// group the transitions within grpwid Hz and return the grouped values
// in grpfrq, grpdif and grplin.

	int strtix;
	double cutoff;

	*grpfrq = 0.0;
	*grpdif = 0.0;
	*grpwgt = 0.0;
	*grplin = 0;
	*grpcnt = 0;

	strtix = *curix;

	if((asncnt > 0) && (*curix > 0) && (*curix <= asncnt))
	{
		cutoff = asnfrq[*curix] + fabs(grpwid);

		while(*curix <= asncnt && asnfrq[*curix] <= cutoff)
		{
			*grpfrq = *grpfrq + asnfrq[*curix];
			*grpdif = *grpdif + fabs(asndif[*curix]);
			*grpwgt = *grpwgt + asnwgt[*curix];
			(*curix)++;
		}
		*grpcnt = *curix - strtix;

		if(*grpcnt > 0)
		{
			*grpfrq = *grpfrq / (double) *grpcnt;
			*grpdif = *grpdif / (double) *grpcnt;
			if(*grpcnt == 1)
				*grplin = asnlin[*curix-1];
			else
				*grplin = asnlin[*curix - *grpcnt/2 -1];
		}
	}
}

#ifdef DUMP_LOG
#undef DUMP_LOG
#endif
