//alloc_spin_space.cpp  allocates global simulation space for SpinWorks

#include <stdio.h>
#include "nummrit.h"
#include "hamfiles.h"
#include "maxvalue.h"
#include "matutil.h"
#include "alloc_spin_space.h"

// /wrkcom/
extern int		*miso43, *le43, *mp43a, *ne43a, **mg, *mtot, *misop, *agp;

// /wrkcom for trnclc/
extern int		**tr_miso43, **tr_le43, **tr_agp, *tr_mga;
extern double	*tr_fdif;

// /mat2cm/
extern int		nm2, km2, ne2, nm32, sym2, wgt2,
				*amaxg, *aming, *ag,
				**fg;

extern double	**bfactf, **bfactg;

extern double	**dh, **da, **db, *d, *dj, *dk;
extern double	*e3, *u3;

extern double   **tub, *t44;

//#define DUMP_LOG

void alloc_spin_space()
{
#ifdef DUMP_LOG
	FILE *fp;
#endif

//allocate space for /mat2cm/ vectors and matrices
#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n\n*** alloc_spin_space(): mat2cm\n");
	fclose(fp);
#endif
	amaxg = ivector(1,MAXGRP);
	aming = ivector(1,MAXGRP);
	ag = ivector(1,MAXGRP);
	fg = imatrix(1, 2, 1, MAXGRP);
	bfactf = matrix(1, MAXF, 1, MAXGRP);
	bfactg = matrix(1, MAXF, 1, MAXGRP);

//allocate for /wrkcom/ vectors and matrices
#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** alloc_spin_space(): wrkcom\n");
	fclose(fp);
#endif
	miso43 = ivector(1, MAXM43);
	le43 = ivector(1, MAXM43);
	mp43a = ivector(1, MAXM43);
	ne43a = ivector(1, MAXM43);
	mg = imatrix(1, MAXGRP, 1, MAXE2);
	mtot = ivector(1, MAXE2);
	misop = ivector(1, MAXE2);
	agp = ivector(1, MAXE2);

// scratch space for trnclc  /wrkcom/
#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** alloc_spin_space(): tr_miso43, tr_le43, tr_agp tr_mga and tr_fdif\n");
	fclose(fp);
#endif
	tr_miso43 = imatrix(1, MAXM43, 1, 2);
	tr_le43 = imatrix(1, MAXM43, 1, 2);
	tr_agp = imatrix(1, MAXE3, 1, 2);	// allocate dynamically in basis with U3
	tr_mga = ivector(1, MAXGRP);		//   and E3 ???
	tr_fdif = vector(1, MAXVAR);

// hamiltonian matrices, eigenvectors, etc. for hamil and trnclc
#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** alloc_spin_space(): allocating space for hamil and trnclc\n");
	fclose(fp);
#endif
	dh = matrix(1, MAXE4, 1, MAXE4);		// for hamil
	da = matrix(1, MAXE4, 1, MAXE4);		// for hamil
	db = matrix(1, MAXE4, 1, MAXE4);		// for hamil
	d  = vector(1, MAXE4);					// for hamil
	dj = vector(1, MAXE4);					// for hamil
	dk = vector(1, MAXE4);					// for hamil

	t44 = vector(1, MAXE4);					// for trnclc
	tub = matrix(1, MAXE4, 1, MAXE4);		// for trnclc	
}

void free_spin_space()
{

#ifdef DUMP_LOG
	FILE *ofp;
#endif

// mat2cm vectors and matrices
#ifdef DUMP_LOG
	ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "*** free_spin_space(): mat2cm\n");
	fclose(ofp);
#endif
	free_ivector(amaxg, 1, MAXGRP);
	free_ivector(aming, 1, MAXGRP);
	free_ivector(ag, 1, MAXGRP);
	free_imatrix(fg, 1, 2, 1, MAXGRP);
	free_matrix(bfactf, 1, MAXF, 1, MAXGRP);
	free_matrix(bfactg, 1, MAXF, 1, MAXGRP);

// wrkcom vectors and matrices
#ifdef DUMP_LOG
	ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "\n*** free_spin_space(): wrkcom\n");
	fclose(ofp);
#endif
	free_ivector(miso43, 1, MAXM43);
	free_ivector(le43, 1, MAXM43);
	free_ivector(mp43a, 1, MAXM43);
	free_ivector(ne43a, 1, MAXM43);
	free_imatrix(mg, 1, MAXGRP, 1, MAXE2);
	free_ivector(mtot, 1, MAXE2);
	free_ivector(misop, 1, MAXE2);
	free_ivector(agp, 1, MAXE2);

// /wrkcom/ for trnclc
#ifdef DUMP_LOG
	ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "\n*** free_spin_space(): tr_miso43, tr_le43, tr_agp tr_mga and tr_fdif\n");
	fclose(ofp);
#endif
	free_imatrix(tr_miso43, 1, MAXM43, 1, 2);
	free_imatrix(tr_le43, 1, MAXM43, 1, 2);
	free_imatrix(tr_agp, 1, MAXE3, 1, 2);		// free dynamically elsewhere??
	free_ivector(tr_mga, 1, MAXGRP);			
	free_vector(tr_fdif, 1, MAXVAR);

// the following are used in hamil() and trnclc()
#ifdef DUMP_LOG
	ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "\n*** free_spin_space(): hamil and trnclc matrices\n");
	fclose(ofp);
#endif
	free_matrix(dh, 1, MAXE4, 1, MAXE4);  
	free_matrix(da, 1, MAXE4, 1, MAXE4);  
	free_matrix(db, 1, MAXE4, 1, MAXE4);  
	free_vector(d, 1, MAXE4);			  
	free_vector(dj, 1, MAXE4);			  
	free_vector(dk, 1, MAXE4);	

	free_vector(t44, 1, MAXE4);
	free_matrix(tub, 1, MAXE4, 1, MAXE4);
}

#ifdef DUMP_LOG
#undef DUMP_LOG
#endif

