// basis generates the basis functions for a NUMARIT type simulation
// numerous "goto's" etc are a holdover from the original FORTRAN

#include <stdio.h>
#include <stdlib.h>
#include "nummrit.h"
#include "hamfiles.h"
#include "maxvalue.h"
#include "matutil.h"
#include "alloc_spin_space.h"
#include <math.h>
#include "basis.h"

extern short sim_debug;

// /isocom/
extern double	*tiiso, *tliso,
				*fliso, *fuiso;
extern int		*isospn, *ntiso,
				*misopa, *misopb, km2p;
extern char		isonam[MXINAM + 1][MAXISO + 1];
extern char		prtiso[MAXISO + 2];

// /grpcom/
extern double	*cshift, **cup, **symcup,
				**dip, **symdip;
extern int		*grpiso, *nucg, *pairg,
				*grpspn, fmaxg[MAXGRP+1][MAXGRP+1], *agpb;
extern short	anydip, symspn;
extern char		**grpnam;

#include "mat2cm.h"

// /spncom/
extern int		ngrp, niso, nvar;

int u3_size;
int e3_size;

// /wrkcom/
int *miso43, *le43, *mp43a, *ne43a, **mg, *mtot, *misop, *agp;

//extern CString sim_out;

//#define DUMP_LOG

///////////////////////////////////////////////////////////////////////////

int basis(const char *sim_out)
{
	int error = FALSE;

	int apack, fvalue, i, iso, j, k, ka, ka2,
		kb, km, km32, kp, kw, l, m, mate2, 
		/* matnum, */ mp, mpack, mpb, mpdif, n, 
		ne3, ne4, ne4b, nm43, nm43a, 
		numerf, numerg, nu3, totdif, syma;

	short  basprt = FALSE;
	short  debug = sim_debug;						 // FALSE for production

	int mxe2=0, mxe3=0, mxe4=0, mxm43=0, mxu3=0,
		cnte2=0, cnte3=0, cnte4=0, cntm43=0, cntu3=0;
	double tote2=0.0, tote3=0.0, tote4=0.0, totm43=0.0, totu3=0.0;

	const int infin = 2147483647;

	// static int **nfgw = imatrix(1, 5, 1, 7); //// <<<<<<
	int **nfgw = imatrix(1, 5, 1, 7);

	FILE *ofp;

#ifdef DUMP_LOG
	FILE *fp;
#endif

	alloc_spin_space();


// allocate space for local variables
#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** basis(): locals\n");
	fclose(fp);
#endif
	int *ktiso = ivector(1, MAXISO);
	int *lgiso = ivector(1, MAXISO);
	int *symg = ivector(1, MAXISO);

	nfgw[1][1]=2;  nfgw[2][1]=1;  nfgw[3][1]=1;  nfgw[4][1]=1; nfgw[5][1]=1;
	nfgw[1][2]=2;  nfgw[2][2]=3;  nfgw[3][2]=1;  nfgw[4][2]=1; nfgw[5][2]=1;
	nfgw[1][3]=5;  nfgw[2][3]=4;  nfgw[3][3]=1;  nfgw[4][3]=1; nfgw[5][3]=1;
	nfgw[1][4]=5;  nfgw[2][4]=9;  nfgw[3][4]=5;  nfgw[4][4]=1; nfgw[5][4]=1;
	nfgw[1][5]=14; nfgw[2][5]=14; nfgw[3][5]=6;  nfgw[4][5]=1; nfgw[5][5]=1;
	nfgw[1][6]=14; nfgw[2][6]=28; nfgw[3][6]=20; nfgw[4][6]=7; nfgw[5][6]=1;
	nfgw[1][7]=42; nfgw[2][7]=48; nfgw[3][7]=27; nfgw[4][7]=8; nfgw[5][7]=1;
//  what does "save" mean in the fortran code

//mx, cnt and tot values initialized at declaration;

//initialize matrices and vectors
	iclear(MAXISO, ntiso, 0);
	matclear(bfactf, 1, MAXF, 1, MAXGRP, 0.0);
	matclear(bfactg, 1, MAXF, 1, MAXGRP, 0.0);

	FILE *scratch1 = fopen("scratch.1", "wb");
	FILE *quant;

// the following is the verbatim translation from the Fortran
// including the God-awful gotos.  Straighten it out when
// possible

	if(debug)quant = fopen("quant_bas.txt", "w");
	if(debug)fprintf(quant, "***** initial basis setup\n");

	u3_size = 0;	e3_size = 0;
	i = 0;			j = 0;  

l210:;
	i = i+1;
	misopa[i] = 0;

l220:;
	j = j+1;
	misopa[i] = misopa[i] + fmaxg[1][j] + fmaxg[2][j];
	if(debug)fprintf(quant, "  misopa[%d] = %d\n", i, misopa[i]);
	lgiso[i] = j;
	if(debug)fprintf(quant, "  lgiso[%d] = %d\n", i, lgiso[i]);
	if(j >= ngrp) goto l240;
	if(grpiso[j] == grpiso[j+1]) goto l220;
	goto l210;

l240:;
	 n = 1;
	 for(i=niso; i>=1; i--)
	 {
		 misopb[i] = n;
		 if(debug)fprintf(quant, "  misopb[%d] = %d\n", i, misopb[i]);
		 j = 1 + 2 * misopa[i];
		 if(n > infin/j) //goto l8010;		//ERROR bail out
		 {
			ofp = fopen(sim_out, "a");
			fprintf(ofp,"SIMULATION: basis(): internal packing overflow\n");
			fclose(ofp);
			error = ERROR;
			return error;
		 }
		 n = n*j;
	 }										// line 250
	 km2p = n;

	 n = 1;
	 for(i=ngrp; i>=1; i--)
	 {
		 agpb[i] = n;
		 if(debug)fprintf(quant, "  agpb[%d] = %d\n", i, agpb[i]);
		 j = 1+fmaxg[1][i] + fmaxg[2][i];
		 if(n > infin/j) //goto l8010;		// ERROR bail out
		 {
			ofp = fopen(sim_out, "a");
			fprintf(ofp, "SIMULATION: basis(): internal packing overflow\n");
			fclose(ofp);
			error = ERROR;
			return error;
		 }
		 n = n*j;
	 }		// line 260

	for(j=1; j<=ngrp; j++)
	{
		for(i=1; i<=2; i++)
		{
			fg[i][j] = fmaxg[i][j];		// line 990
		}
		if(debug)fprintf(quant, "  fg[1][%d] = %d      fg[2][%d] = %d\n",
			     j, fg[1][j], j, fg[2][j]);
	}

	if(debug)fprintf(quant, "****  finished dump of initial setup\n");
	if(debug)fclose(quant);

	nm2 = 0;
	goto l1050;

l1000:;		// continue

	for(j=1; j<=ngrp; j++)
	{
		if(grpspn[j] <= 1)
		{
			for(i=1; i<=2; i++)
			{
				fg[i][j] = fg[i][j] - 2;
				if(fg[i][j] >= 0) goto l1050;
				fg[i][j] = fmaxg[i][j];
			}							// line 1020
		}								// endif
	}									// line 1030

	if(nm2 == 0)
	{
		ofp = fopen(sim_out, "a");
		fprintf(ofp,"SIMULATION: basis(): ERROR: nm2 = 0\n  check your input\n\n");
		fclose(ofp);
		error = ERROR;
		return error;
	}

	if(km2p > infin/nm2) //goto l8010;	// error! bail out
	{
		ofp = fopen(sim_out, "a");
		fprintf(ofp,"SIMULATION: basis(): internal packing overflow\n");
		fclose(ofp);
		error = ERROR;
		return error;
	}
	goto l2000;

l1050:;	// continue

	sym2 = 0;
	kw = 1;

	if(symspn == FALSE) goto l1100;
	
	for(i=1; i<=ngrp; i++)
	{
		if(pairg[i] == 2)
		{
			fvalue = fg[1][i] - fg[2][i];
			if(fvalue > 0) goto l1000;
			if(fvalue < 0)
			{
				kw = 2;
				goto l1100;
			}							// endif
		}								// endif
	}									// line 1070 (continue)
	sym2 = 1;

l1100:;	// continue

	for(j=1; j<=ngrp; j++)
	{
		if(grpspn[j] <= 1)
		{
			for(i=1; i<=2; i++)
			{
				l = fmaxg[i][j] -2;
				if(l > 0)
				{
					m = 1 + ((fg[i][j])/2);
					kw = kw * nfgw[m][l];
				}						// endif
			}							// line 1120 (continue)
		}								// endif
	}									// line 1130 (continue)

	wgt2 = kw;

	for(i=1; i<=ngrp; i++)
	{
		amaxg[i] = fg[1][i] + fg[2][i];
		aming[i] = abs(fg[1][i] - fg[2][i]);
		ag[i] = amaxg[i];
		if(pairg[i] != 2) 
			symg[i] = 0;
		else
			symg[i] = 1;
	}									// line 1150 (continue)


	for(i=1; i<=ngrp; i++)
	{
		if(pairg[i] == 2)
		{
			numerg = (fg[1][i] * (fg[1][i] + 2))
				   - (fg[2][i] * (fg[2][i] + 2));
			k = amaxg[i];
			kp = k + 2;
			l = aming[i];
			for(j=1; j<=MAXF; j++)
			{
				bfactf[j][i] = 0.0;
				bfactg[j][i] = 0.0;
				ka = 2 * j;
				if(!(ka > k || ka < l))
				{
					ka2 = ka * ka;		// necessary???
					numerf = (ka2 - l*l) * (kp*kp -ka2);
					bfactf[j][i] = (double)numerf/(double)(4*ka2*(4*ka2-4));
					bfactg[j][i] = (double)numerg/(double)(ka*(ka+2));
				}						// endif
			}							// line 1170 (continue)
		}								// endif
	}									// line 1180 (continue)

l1190:; // continue	

	ne2 = 0;
	goto l1230;

l1200:; // continue

	for(i=1; i<=ngrp; i++)
	{
		if(pairg[i] == 2)
		{
			ag[i] = ag[i] - 2;
			if(ag[i] >= aming[i]) goto l1220;
			ag[i] = amaxg[i];
			symg[i] = 1;
		}								// endif
	}									// line 1210 (continue)

	goto l1500;
l1220:;	// continue

	symg[i] = -symg[i];

l1230:; // continue

	if(sym2 != 0)
	{
		syma = 1;
		for(i=1; i<=ngrp; i++)
		{
			if(symg[i] == -1) syma = -syma;
		}								// line 1240 (continue)
		if(syma != sym2) goto l1200;
	}									// endif

	n = ne2 + 1;

	k = 1;
	for(i=1; i<=ngrp; i++)
	{
		k = k * (ag[i]+1);
	}									// line 1260 (continue)

	mate2 = n + k;

	if(mate2 > mxe2) mxe2 = mate2;

	tote2 = tote2 + mate2;
	cnte2 = cnte2 + 1;
	if(basprt == TRUE)					// print basis info
	{
		ofp = fopen(sim_out, "a");
		fprintf(ofp, "SIMULATION: basis(): size of mat2 = %d\n",
			mate2);
		fclose(ofp);
	}									// endif

	if(mate2 > MAXE2) // goto l8020;	// disaster! bail out
	{
		ofp = fopen(sim_out, "a");
		fprintf(ofp,"SIMULATION: basis(): limit is  %d levels in a mat(2)\n",
			     MAXE2);
		fclose(ofp);
		error = ERROR; 
		return error;
	}

 
	apack = 0;
	for(i=1; i<=ngrp; i++)
	{
		apack = apack + ag[i] * agpb[i];
		mg[i][n] = ag[i];
	}									// line 1280 (return)

l1300:;  // continue

	ne2 = n;
	mpack = 0;
	mtot[ne2] = 0;
	k = 1;
	for(iso = 1; iso <= niso; iso++)
	{
		l = lgiso[iso];
		m = 0;
		for(j=k; j<=l; j++)
		{
			m = m + mg[j][ne2];
		}								// line 1310 (continue)
		mpack = mpack + misopb[iso] * (m + misopa[iso]);
		mtot[ne2] = mtot[ne2] + m;
		k = l+1;
	}									// line 1320 (continue)

	misop[ne2] = mpack;
	agp[ne2] = apack;
	n = ne2 + 1;
	for(i=1; i<=ngrp; i++)
	{
		mg[i][n] = mg[i][ne2];
	}									// line 1330 (continue)

	for(km=1; km<=ngrp; km++)
	{
		mg[km][n] = mg[km][n] - 2;
		if((mg[km][n] + ag[km]) >= 0) goto l1300; // barf!
		mg[km][n] = ag[km];
	}									// line 1340 (continue)

	goto l1200;

l1500:;	// continue

	  nm32 = 0;
	  l = -10000;
	  kb = 1;
	  for(ka = 1; ka <= ne2; ka++)
	  {
		  m = -10000;
		  for(i = ka; i <= ne2; i++)
		  {
			  totdif = m - mtot[i];
			  if((totdif < 0) || 
				 (totdif == 0 && misop[kb] < misop[i]))
			  {
				  kb = i;
				  m = mtot[i];
			  }							// endif
		  }								// line 1540 (continue)
		  if(l != m) nm32 = nm32 +1;
		  l = m;

		  if(kb != ka) 
		  {
				for(i=1; i<=ngrp; i++)
				{
					m = mg[i][ka];
					mg[i][ka] = mg[i][kb];
					mg[i][kb] = m;
				}						// line 1550 (continue)
				m = mtot[ka];
				mtot[ka] = mtot[kb];
				mtot[kb] = m;
				m = agp[ka];
				agp[ka] = agp[kb];
				agp[kb] = m;
				m = misop[ka];
				misop[ka] = misop[kb];
				misop[kb] = m;
		  }								// endif
	  }									// line 1560 (continue)


	if(nm32 >=2)
	{
		if(debug)
		{
			quant = fopen("quant_bas.txt", "a");
			if(debug)fprintf(quant, "*** starting scratch1 file write\n");
			fflush(quant);
		}
		nm2 = nm2 +1;

		fwrite(&nm2, sizeof(int), 1, scratch1);
		fwrite(&ne2, sizeof(int), 1, scratch1);
		fwrite(&nm32, sizeof(int), 1, scratch1);
		fwrite(&wgt2, sizeof(int), 1, scratch1);
		fwrite(&sym2, sizeof(int), 1, scratch1);
		if(debug)
		{
			fprintf(quant, "       nm2, ne2, nm32, wgt2 and sym2 written\n");
			fflush(quant);
		}

//		fwrite(&amaxg[1], sizeof(int), MAXGRP, scratch1);
//		fwrite(&aming[1], sizeof(int), MAXGRP, scratch1);

		for(i=1; i <= MAXGRP; i++)
			fwrite(&amaxg[i], sizeof(int), 1, scratch1);

		for(i=1; i <= MAXGRP; i++)
			fwrite(&aming[i], sizeof(int), 1, scratch1);

		if(debug)
		{
			fprintf(quant, "       amaxg[] and aming[] written\n");
			fflush(quant);
		}

//		fwrite(&fg[1][1], sizeof(int), 2*MAXGRP, scratch1);
		for(i = 1; i <= MAXGRP; i++)fwrite(&fg[1][i], sizeof(int), 1, scratch1);
		for(i = 1; i <= MAXGRP; i++)fwrite(&fg[2][i], sizeof(int), 1, scratch1);
		if(debug)
		{
			fprintf(quant, "       fg[][] written\n");
			fflush(quant);
		}

//		fwrite(&bfactf[1][1], sizeof(int), MAXF*MAXGRP, scratch1);
		for(i = 1; i <= MAXF; i++)
			for(j = 1; j <= MAXGRP; j++)
				fwrite(&bfactf[i][j], sizeof(int), 1, scratch1);

//		fwrite(&bfactg[1][1], sizeof(int), MAXF*MAXGRP, scratch1);   //  quits here???
		for(i = 1; i <= MAXF; i++)
			for(j = 1; j <= MAXGRP; j++)
				fwrite(&bfactg[i][j], sizeof(int), 1, scratch1);

		if(debug)
		{
			fprintf(quant, "       bfactf[] and bfactg[] written\n");
			fflush(quant);
		}

		if(debug)
		{
			fprintf(quant, " nm2 = %d   ne2 = %d    nm32 = %d   wgt2 = %d   sym2 = %d\n",
			nm2, ne2, nm32, wgt2, sym2);
			fprintf(quant, "\n\n");
			fclose(quant);
		}

		mtot[ne2+1] = -10000;
		misop[ne2+1] = -10000;
		k = 1;
		mp = misop[k];
		ne4 = 1; 
		km32 = 0;

l1610:; // continue

		ne3 = 0;
		nu3 = 0;
		nm43 = 0;
		m = mtot[k];
		ka = k;

l1620:; // continue

		ne3 = ne3 +1;
		k = k + 1;


		if(mp != misop[k])
		{
			nm43 = nm43 +1;

//			MXM43  = MAX( MXM43, NM43 )		// check
			if(nm43 > mxm43) mxm43 = nm43;

			totm43 = totm43 + nm43;
			cntm43 = cntm43 +1;
			if(basprt == TRUE)
			{
				// print basis information
			}								// endif
		
			if(nm43 > MAXM43) goto l8030;	// bail out

		 // mxe4 = fmax(mxe4, ne4);			// check!!!
			if(ne4 > mxe4) mxe4 = ne4;

			tote4 = tote4 + ne4;
			cnte4 = cnte4 + 1;
			if(basprt == TRUE)
			{
				// print basis information
			}

			if(ne4 > MAXE4) goto l8040;		// bail out
			le43[nm43] = ne3;
			nu3 = nu3 + ne4*ne4;
			miso43[nm43] = mp;
			mp = misop[k];
			ne4 = 0;
		}									// endif   
		ne4 = ne4 +1;
		if (m == mtot[k]) goto l1620;		// gadzooks!!

	 // mxu3 = fmax(mxu3, nu3);				// check!!
		if(nu3 > mxu3) mxu3 = nu3;

		totu3 = totu3 + nu3;
		u3_size = u3_size + nu3;
		cntu3 = cntu3 +1;
		if(basprt == TRUE)
		{
			// print basis info
		}

		if(basprt)
		{
			ofp = fopen(sim_out, "a");
			fprintf(ofp, "***  Size of U3 (eigenvalue matrix) = %d\n", nu3);
			fclose(ofp);
		}

		if(nu3 > MAXU3) goto l8050;			// bail out!!

	 // mxe3 = fmax(mxe3, ne3);				// check!!
		if(ne3 > mxe3) mxe3 = ne3;

		tote3 = tote3 + ne3;
		e3_size = e3_size + ne3;
		cnte3++;

		if(basprt == TRUE)					// print basis information
		{
			ofp = fopen(sim_out, "a");
			fprintf(ofp, "***  Number of basis functions in matrix(3) = %d\n", ne3);
			fclose(ofp);
		}
	
		if(ne3 > MAXE3)							// bail out!!
		{
			fclose(scratch1);
			ofp = fopen(sim_out, "a");
			fprintf(ofp, "***  Too many basis functions! simulation exiting\n");
			fclose(ofp);
			return ERROR;
		}

		km32 = km32 + 1;
		kb = k - 1;

// writing to "QUANT" (scratch1)
		int n_obj = 0;
		n_obj = n_obj + fwrite(&km32, sizeof(int), 1, scratch1);
		n_obj = n_obj + fwrite(&m, sizeof(int), 1, scratch1);
		n_obj = n_obj + fwrite(&nm43, sizeof(int), 1, scratch1);
		n_obj = n_obj + fwrite(&ne3, sizeof(int), 1, scratch1);

		for(i=1; i<=nm43; i++)
		{
			n_obj = n_obj + fwrite(&le43[i], sizeof(int), 1, scratch1);
			n_obj = n_obj + fwrite(&miso43[i], sizeof(int), 1, scratch1);
		}

		for(j=ka; j<=kb; j++)
			n_obj = n_obj + fwrite(&agp[j], sizeof(int), 1, scratch1);

		for(i=1; i<=ngrp; i++)
		{
			for(j=ka; j<=kb; j++)
			{
				n_obj = n_obj + fwrite(&mg[i][j], sizeof(int), 1, scratch1);
			}
		}
		

		if(debug)
		{
			quant = fopen("quant_bas.txt", "a");
			fprintf(quant, "BASIS: writing QUANT (scratch1)\n");
			fprintf(quant, "  km32= %d, m= %d, nm43= %d, ne3= %d\n",
				km32, m, nm43, ne3);

			int ii;
			for(ii=1; ii<=nm43; ii++)
			{
				fprintf(quant, "  le43[%d]=%d, miso43[%d]=%d;  ",ii, le43[ii], ii, miso43[ii]);
			}
			fprintf(quant, "\n");

			for(ii=ka; ii<=kb; ii++)
			{
				fprintf(quant, "  agp[%d]=%d",ii, agp[ii]);
			}
			fprintf(quant,"\n");

			for(ii=1; ii<=ngrp; ii++)
			{
				int jj;
				for(jj=ka; jj<=kb; jj++)
				{
					fprintf(quant, "  mg[%d][%d]=%d",ii, jj, mg[ii][jj]);
				}
			fprintf(quant,"\n");
			}

			fprintf(quant, "scratch1 objects written = %d\n\n",n_obj);
			fclose(quant);
		}
	 

		if(km32 != 1)
		{
			j = 0;
			for(kb = 1; kb <= nm43; kb++)
			{
				ne4b = le43[kb] - j;
				j = le43[kb];
				mpb = miso43[kb];
				for(iso = 1; iso <= niso; iso++)
				{
					mpdif = 2*misopb[iso];
					for(ka = 1; ka <= nm43a; ka++)
					{
						if(mp43a[ka] - mpb == mpdif)
						{
							ntiso[iso] = ntiso[iso] + ne43a[ka]*ne4b;
						}					// endif
					}						// line 1660 (continue)
				}							// line 1670 (continue)
			}								// line 1670 (continue)
		}									// endif

		j = 0;
		nm43a = nm43;
		for(ka = 1; ka<=nm43a; ka++)
		{
			ne43a[ka] = le43[ka] - j;
			j = le43[ka];
			mp43a[ka] = miso43[ka];
		}									// line 1690 (continue)

		if(km32 < nm32) goto l1610;			// barf!!


	}	// end of a mat2					// endif
	
	if(sym2 != 1) goto l1000;
	sym2 = -1;
	goto l1190;

l2000:;	// continue
 
	n = 0;
	ka = 1;
	for(iso = 1; iso <= niso; iso ++)
	{
		m = n + 1;
		n = lgiso[iso];
		i = isospn[iso];
		kb = 0;
		for(k = m; k<=n; k++)
		{
			ka = ka*((int)pow((double)(i+1), (nucg[k]*pairg[k])));	// check ////!!!!!!
			kb = kb + nucg[k]*i*(i+2)*pairg[k];
		}									// line 2030 (continue)
		ktiso[iso] = kb;
	}										// line 2050 (continue)
 

	for(iso=1; iso<=niso; iso++)
	{
		ktiso[iso] = (ktiso[iso]*ka)/6;
		tiiso[iso] = (double)ktiso[iso];
	}										// line 2070 (continue)

	ofp = fopen(sim_out, "a");
	fprintf(ofp, "\nBasis functions generated, transition information:\n\n");
	for(iso=1; iso<=niso; iso++)		
		fprintf(ofp,"   Species %2d,  up to %7d transitions, intensity %7d\n",
		            iso, ntiso[iso], ktiso[iso]);
	fprintf(ofp, "\n   Total size of U3 matrix = %d double words\n", u3_size);
	fprintf(ofp, "   Total sixe of E3 matrix = %d double words\n", e3_size);
	fclose(ofp);

// these are allocated here, to prevent them being allocated and deallocated
// in hamil() during an iteration


	u3 = vector(1, u3_size + 1);			// for hamil
	e3 = vector(1, e3_size + 1);			// for hamil
	tr_u3 = matrix(1, u3_size + 1, 1, 2);	// for trnclc
	tr_e3 = matrix(1, e3_size + 1, 1, 2);	// for trnclc

	// free space used by local variables
#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** basis(): freeing locals\n");
	fclose(fp);
#endif
	free_ivector(ktiso, 1, MAXISO);
	free_ivector(lgiso, 1, MAXISO);
	free_ivector(symg, 1, MAXISO);


	fclose(scratch1);
	return error;

l8030:
	fclose(scratch1);
	return ERROR;
l8040:
	fclose(scratch1);
	return ERROR;
l8050:
	fclose(scratch1);
	return ERROR;
}

#ifdef DUMP_LOG
#undef DUMP_LOG
#endif

