// finish_sim.cpp
// Clean-up and release space after a NUMARIT type simulation.
// No reference to Reino L. or Perch in the file name!

#include <stdio.h>
#include <unistd.h>
#include "nummrit.h"
#include "maxvalue.h"
#include "matutil.h"
#include "alloc_spin_space.h"

// /grpcom/
extern double	*cshift, **cup, **symcup,
				**dip, **symdip;
extern int		*grpiso, *nucg, *pairg,
				*grpspn, fmaxg[MAXGRP+1][MAXGRP+1], *agpb;
extern short	anydip, symspn;
extern char		**grpnam;

// /mat2cm/
extern int		nm2, km2, ne2, nm32, sym2, wgt2,
				*amaxg, *aming, *ag,
				**fg;
extern double	**bfactf, **bfactg;
extern double	**dh;
extern double	**da, **db;
extern double	*d, *dj, *dk;
extern double	*u3, *e3;
extern double	**tr_u3, **tr_e3;
extern double   **tub, *t44;

// /isocom/
extern double	*tiiso, *tliso,
				*fliso, *fuiso;
extern int		*isospn, *ntiso,
				*misopa, *misopb, km2p;

// /prmcom/
extern int		prmsiz, prmnum, prmi, 
				prmj, prmd, lowvar;
extern double	rngdef, rngnit, rngmin;
extern double	*asnhi, *asnlow,
				*rhics,	*rlocs, rngfra,
				**rhiccj, **rloccj,
				**rhiscj, **rloscj,
				**rhiccd, **rloccd,
				**rhiscd, **rloscd;
extern int		*asnmin, *asnnum,
				**prmmap, *ivcs,
				**ivcup, **ivdip;

// /asncom/
extern double	*asnfrq, *asnwgt, *asndif, *asnclc, asnint;
extern int		asncnt, *asnlin;

// /obscom/
extern int		obscnt, *obslin;
extern double	*obsfrq, *obsint;

// common block /wrkcom/
extern int		*miso43, *le43, *mp43a, *ne43a, **mg, 
				*mtot, *misop, *agp;

extern int u3_size;
extern int e3_size;

//#define DUMP_LOG

int finish()
{
// tidies-up and frees global simulation space
	int error = FALSE;

#ifdef DUMP_LOG
	FILE *ofp;
#endif

free_spin_space();

// allocated in basis()
	free_vector(u3, 1, u3_size + 1);
	free_vector(e3, 1, e3_size + 1);
	free_matrix(tr_u3, 1, u3_size + 1, 1, 2);	// for trnclc
	free_matrix(tr_e3, 1, e3_size + 1, 1, 2);	// for trnclc

// /grpcom/
#ifdef DUMP_LOG
	ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "\n\n*** finish(): grpcom\n");
	fclose(ofp);
#endif
	free_vector(cshift, 1, MAXGRP);
	free_matrix(cup, 1, MAXGRP, 1, MAXGRP);
	free_matrix(symcup, 1, MAXGRP, 1, MAXGRP);
	free_matrix(dip, 1,  MAXGRP, 1, MAXGRP);
	free_matrix(symdip, 1, MAXGRP, 1, MAXGRP);
	free_ivector(grpiso, 1, MAXGRP);
	free_ivector(nucg, 1, MAXGRP);
	free_ivector(pairg, 1, MAXGRP);
	free_ivector(grpspn, 1, MAXGRP);
//	free_imatrix(fmaxg, 1, MAXGRP, 1, MAXGRP);
	free_ivector(agpb, 1, MAXGRP);

// /isocom/
#ifdef DUMP_LOG
	ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "\n*** finish(): isocom\n");
	fclose(ofp);
#endif
	free_vector(tiiso, 1, MAXISO);
	free_vector(tliso, 1, MAXISO);
	free_vector(fliso, 1, MAXISO);
	free_vector(fuiso, 1, MAXISO);
	free_ivector(isospn, 1, MAXISO);
	free_ivector(ntiso, 1, MAXISO);
	free_ivector(misopa, 1, MAXISO);
	free_ivector(misopb, 1, MAXISO);

// /prmcom/
#ifdef DUMP_LOG
	ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "\n*** finish(): prmcom\n");
	fclose(ofp);
#endif
	free_vector(asnhi, 1, MAXGRP);
	free_vector(asnlow, 1, MAXGRP);
	free_vector(rhics, 1, MAXGRP);
	free_vector(rlocs, 1, MAXGRP);
	free_matrix(rhiccj, 1, MAXGRP, 1, MAXGRP);
	free_matrix(rloccj, 1, MAXGRP, 1, MAXGRP);
	free_matrix(rhiscj, 1, MAXGRP, 1, MAXGRP);
	free_matrix(rloscj, 1, MAXGRP, 1, MAXGRP);
	free_matrix(rhiccd, 1, MAXGRP, 1, MAXGRP);
	free_matrix(rloccd, 1, MAXGRP, 1, MAXGRP);
	free_matrix(rhiscd, 1, MAXGRP, 1, MAXGRP);
	free_matrix(rloscd, 1, MAXGRP, 1, MAXGRP);
	free_ivector(asnmin, 1, MAXGRP);
	free_ivector(asnnum, 1, MAXGRP);
	free_imatrix(prmmap, 1, 4, 0, MAXVAR);
	free_ivector(ivcs, 1, MAXGRP);
	free_imatrix(ivcup, 1, MAXGRP, 1, MAXGRP);
	free_imatrix(ivdip, 1, MAXGRP, 1, MAXGRP);

// asncom
#ifdef DUMP_LOG
	ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "\n*** finish(): asncom\n");
	fclose(ofp);
#endif
	 free_ivector(asnlin, 1, MAXASN);
	 free_vector(asnfrq, 1, MAXASN);
	 free_vector(asnwgt, 1, MAXASN);
	 free_vector(asnclc, 1, MAXASN);
	 free_vector(asndif, 1, MAXASN);

// obscom
#ifdef DUMP_LOG
	ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "\n*** finish(): obscom\n");
	fclose(ofp);
#endif
	 free_ivector(obslin, 1, MAXOBS);
	 free_vector(obsfrq, 1, MAXOBS);
	 free_vector(obsint, 1, MAXOBS);

// clear scratch files
	 unlink("scratch.1");
	 unlink("scratch.2");
	 unlink("scratch.3");
	 unlink("scratch.4");
	 unlink("isoscr.1");
	 unlink("isoscr.2");
	 unlink("isoscr.3");
	 unlink("isoscr.4");

	return error;
}

#ifdef DUMP_LOG
#undef DUMP_LOG
#endif
