// Generate and diagonalize the Hamiltonian matricies for a NUMARIT type
// simulation

#include <stdio.h>
#include <stdlib.h>
#include "nummrit.h"
#include "maxvalue.h"
#include "matutil.h"
#include "matrix_diag.h"
#include "hamil.h"
#include <math.h>

extern short sim_debug;
// /isocom/
extern double	*tiiso, *tliso,
				*fliso, *fuiso;
extern int		*isospn, *ntiso,
				*misopa, *misopb, km2p;
extern char		isonam[MXINAM + 1][MAXISO + 1];
extern char		prtiso[MAXISO + 2];

// /grpcom/
extern double	*cshift, **cup, **symcup,
				**dip, **symdip;
extern int		*grpiso, *nucg, *pairg,
				*grpspn, fmaxg[MAXGRP+1][MAXGRP+1], *agpb;
extern short	anydip, symspn;
extern char		**grpnam;

// /mat2cm/
extern int		nm2, km2, ne2, nm32, sym2, wgt2,
				*amaxg, *aming, *ag,
				**fg;
extern double	**dh;
extern double	**da, **db;
extern double	*d, *dj, *dk;
extern double	*e3, *u3;

// /spncom/
extern int		ngrp, niso, nvar;

// /flgcom/
extern short	autasn, autdel, optmiz, prtptl,
				prtkey, prtdif, prtvcm, pracnd;

// /prmcom/
extern int		prmsiz, prmnum, prmi, 
				prmj, prmd, lowvar;
extern double	rngdef, rngnit, rngmin;
extern double	*asnhi, *asnlow,
				*rhics,	*rlocs, rngfra,
				**rhiccj, **rloccj,
				**rhiscj, **rloscj,
				**rhiccd, **rloccd,
				**rhiscd, **rloscd;
extern int		*asnmin, *asnnum,
				**prmmap, *ivcs,
				**ivcup, **ivdip;

extern double	**bfactf, **bfactg;

// /wrkcom/
extern int *miso43, *le43, *mp43a, *ne43a, **mg, *mtot, *misop, *agp;

extern int u3_size;		// U3 matrix size calculated in basis()
extern int e3_size;		// E3 matrix size calculated in basis();

//extern CString sim_out;

//#define DUMP_LOG

/////////////////////////////////////////////////////////////////////////////////
// assemble and diagonalize the hamiltonian matricies
/////////////////////////////////////////////////////////////////////////////////


int hamil(const char *sim_out, short iting, short dodifs)
{
	int		error = FALSE;
	short	sortit = TRUE; short sortup = FALSE;
	short	debug = sim_debug;							// FALSE for production
	double	hh, sgnval, cuptmp, fcttmp,
			dfactq, dfactr, dfacts, bfact;
	double	*shift, **cupind, **cupdip;
	int		*agj;
	int		aa, ai, apj, aq, ar, i, ii, iv,
			j, jj, jv, k, kdif, kdifa, kdifm, kg,
			km32, km43, kq, kr, k1, k2, l, m, 
			mm, mmq, mmr, mq, mr, n, ne3, ne4,
			nm43, nu3, nw;

	double  *pd, **pdint;
	int		*iiint, *jjint;

#ifdef DUMP_LOG
	FILE	*fp;
#endif

//	double	*d, *dj, *dk;				// now allocated in basis()
//	double	**dh;						// repeated allocation/deallocation 
//	double	**da, **db;					// here caused heap fragmentation
//	double	*e3, *u3;

//	int     *miso43, *le43, *mp43a, *ne43a, **mg, *mtot, *misop, *agp;

	int u3_size_local = 0;

//     If we are doing an optimization, for each variable parameter
//     we must calculate the partial derivative of the involved
//     hamiltonian matrix element with respect to the parameter. The
//     derivative is placed in the appropriate place in vector PD.
//     This is done only for non-zero elements. It need only be done
//     once for each case (at the zero'th iteration) since the
//     derivatives are independent of the parameter values. These are
//     calculated for all parameters, even those which don't have
//     enough transitions assigned (ie, don't examine PRMMAP)!
//
//     The following variables are used in derivative calculation:
//       pd, pdint, iiint, jjint, d, da, db, nw, jv

// allocate some space

#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n\n*** hamil(): wrkcom\n");
	fclose(fp);
#endif
	shift = vector(1, MAXGRP);
	cupind = matrix(1, MAXGRP, 1, MAXGRP);
	cupdip = matrix(1, MAXGRP, 1, MAXGRP);
	agj = ivector(1, MAXGRP);

	pd = vector(1, MAXVAR);
	pdint = matrix(1, MAXVAR, 1, MPDTRI);
	iiint = ivector(1, MPDTRI);
	jjint = ivector(1, MPDTRI);

//	dh = matrix(1, MAXE4, 1, MAXE4);		// dh, da and db are 256 * 256 matrices!
//	dj = vector(1, MAXE4);					// for efficiency, now allocated in basis()
//	dk = vector(1, MAXE4);
//	da = matrix(1, MAXE4, 1, MAXE4);
//	db = matrix(1, MAXE4, 1, MAXE4);
//	d = vector(1, MAXE4);

//	u3 = vector(1, u3_size+1);
//	e3 = vector(1, e3_size+1);

	FILE *scratch1 = fopen("scratch.1", "rb");	// QUANT
	FILE *scratch2 = fopen("scratch.2", "wb");	// EU

	FILE *scratch3;
	FILE *scratch4;
	FILE *quant;
	FILE *energ;
	FILE *ofp = fopen(sim_out, "a");

	if(dodifs == TRUE)
	{
//		scratch4 = fopen("scratch.4", "w+b");	// IDIF
		scratch3 = fopen("scratch.3", "wb");	// IDERIV
		rewind(scratch3);						// Why is this necessary ??????
//		rewind(scratch4);
	}


	for(i=1; i<=ngrp; i++)
	{
		shift[i] = cshift[i];
		for(j=1; j<=ngrp; j++)
		{
			cupind[i][j] = symcup[i][j];
			cupdip[i][j] = symdip[i][j];
		}
	}
	apj = -1;

	prtkey = TRUE;

	if(debug)quant = fopen("quant_hamil.txt", "w");
	if(debug)energ = fopen("energies.txt","w");

l1000:;	// continue

	fread(&km2, sizeof(int), 1, scratch1);
	fread(&ne2, sizeof(int), 1, scratch1);
	fread(&nm32, sizeof(int), 1, scratch1);
	fread(&wgt2, sizeof(int), 1, scratch1);
	fread(&sym2, sizeof(int), 1, scratch1);

//	fread(&amaxg[1], sizeof(int), MAXGRP, scratch1);
//	fread(&aming[1], sizeof(int), MAXGRP, scratch1);
	for(i=1; i <= MAXGRP; i++)
		fread(&amaxg[i], sizeof(int), 1, scratch1);
	for(i=1; i <= MAXGRP; i++)
		fread(&aming[i], sizeof(int), 1, scratch1);

//	fread(&fg[1][1], sizeof(int), 2*MAXGRP, scratch1);
	for(i = 1; i <= MAXGRP; i++) fread(&fg[1][i], sizeof(int), 1, scratch1);
	for(i = 1; i <= MAXGRP; i++) fread(&fg[2][i], sizeof(int), 1, scratch1);

//	fread(&bfactf[1][1], sizeof(int), MAXF*MAXGRP, scratch1);
	for(i = 1; i <= MAXF; i++)
		for(j = 1; j <= MAXGRP; j++)
			fread(&bfactf[i][j], sizeof(int), 1, scratch1);

//	fread(&bfactg[1][1], sizeof(int), MAXF*MAXGRP, scratch1);
	for(i = 1; i <= MAXF; i++)
		for(j = 1; j <= MAXGRP; j++)
			fread(&bfactg[i][j], sizeof(int), 1, scratch1);

	if(debug)fprintf(quant, "HAMIL: reading quant\n");
	if(debug)fprintf(quant, "   km2= %d  ne2= %d, nm32= %d, wgt2= %d, sym2= %d\n",
		km2, ne2, nm32, wgt2, sym2);


	if(prtkey && !iting)
	{
		if(debug)
		{
			fprintf(energ, "********************************\n\n");
			fprintf(energ, " Mat(2): %3d", km2);
			fprintf(energ, " Weight: %3d", wgt2);
			if(sym2 < 0)
				fprintf(energ, " Anti");
			else if(sym2 == 0)
				fprintf(energ, " Not ");
			else	
				fprintf(energ, "     ");
			fprintf(energ, " Symmetric\n");
		}

		k = 1;
		int temp_spin[20];

		for(i = 0; i < 20; i++) temp_spin[i] = 0;

		for(i=1; i <= ngrp; i++)								// do 1050
		{
			k++;
			temp_spin[k] = fg[1][i];
			if(pairg[i] == 2) temp_spin[k+MAXGRP] = fg[2][i];
		}														// 1050

		if(debug)
		{
			fprintf(energ, " Group F Values (*2):");
			for(i = 2; i <= ngrp+1; i++)
				fprintf(energ, " %d ", temp_spin[i]);
			fprintf(energ, "                     ");
			for( i = MAXGRP+2; i <= MAXGRP+ngrp+1; i++)
				fprintf(energ, " %d ", temp_spin[i]);
			fprintf(energ, "\n ");
			fprintf(energ, "  Matrix(3) Matrix(4) Isotopic species  Z-spins E-levels\n");
		}
	}			// endif

l1100:; // continue

	fread(&km32, sizeof(int), 1, scratch1);
	fread(&m, sizeof(int), 1, scratch1);
	fread(&nm43, sizeof(int), 1, scratch1);
	fread(&ne3, sizeof(int), 1, scratch1);

	for(i=1; i<=nm43; i++)
	{
		fread(&le43[i], sizeof(int), 1, scratch1);
		fread(&miso43[i], sizeof(int), 1, scratch1);
	}

	for(j=1; j<=ne3; j++)
		fread(&agp[j], sizeof(int), 1, scratch1);

	for(i=1; i<=ngrp; i++)
	{
		for(j=1; j<=ne3; j++)
		{
			fread(&mg[i][j], sizeof(int), 1, scratch1);
		}
	}

	if(debug)
	{
		fprintf(quant, "HAMIL: read QUANT (scratch1)\n");
		fprintf(quant, "  km32= %d, m= %d, nm43= %d, ne3= %d\n",
				km32, m, nm43, ne3);

		for(i=1; i<=nm43; i++)
				fprintf(quant, "  le43[%d]=%d, miso43[%d]=%d;  ",i, le43[i], i, miso43[i]);

		fprintf(quant, "\n");

		for(i=1; i<=ne3; i++)
			fprintf(quant, "  agp[%d]=%d",i, agp[i]);

		fprintf(quant,"\n");

		for(i=1; i<=ngrp; i++)
		{
			for(int j=1; j<=ne3; j++)
			{
				fprintf(quant, "  mg[%d][%d]=%d", i, j, mg[i][j]);
			}
			fprintf(quant,"\n");
		}
	}
	

	nu3 = 0;
	km43 = 1;
	n = 0;

	if(dodifs == TRUE)
	{
		fwrite(&ne3, sizeof(int), 1, scratch3);   /// something wrong here?
	}

	if(prtkey && !iting && debug)		// new mat(3) heading
		fprintf(energ, "  mat(3) = %d", km32);

l1200:;	// continue

// start of a matrix(4)

// Throughout NUMMR, all quantum numbers are represented
// by integers twice their true values. Constants
// appearing in any expression involving spin quantum
// numbers must also be doubled. If higher powers of
// quantum numbers are used, constants become higher
// powers of 2.
//
// The indices i and j number basis functions within
// this matrix(3). ii and jj number the same basis
// functions within this matrix(4). Suffixes I and J on
// other quantities also refer to particular basis
// functions. Suffixes q and r indicate particular
// groups or pairs within a basis function. m and n are
// the first and last basis functions in this matrix(4).

	m = n + 1;
	n = le43[km43];
	ne4 = n - m + 1;

	jj = 0;
	nw = 0;

	for(j=m; j<=n; j++)						// do 2000 (big loop!)
	{
		jj = jj + 1;

		if(apj != agp[j])
		{
			apj = agp[j];
			aa = apj;
			for(kg = 1; kg<=ngrp; kg++)
			{
				agj[kg] = aa/agpb[kg];
				aa = aa - agj[kg]*agpb[kg];
			}								// line 1210 (continue)
		}									// endif
		ii = 0;
		for(i=m; i<=j; i++)
		{
			ii = ii + 1;
			if(dodifs == TRUE)
			{
				for(jv=1; jv<=MAXVAR; jv++) pd[jv] = 0.0;
			}								// endif

			hh = 0.0;

			if(i != j)	// off-diagonal
			{
				if(agp[i] == agp[j])
				{
					kdif = -2;

					for(kg = 1; kg <=ngrp; kg++) // do 1330
					{
						mm = mg[kg][i] - mg[kg][j];
						if(mm != 0)
						{
							if(abs(mm) > 2) goto l1900;
							kdif = kdif + 1;
							if(kdif > 0) goto l1900;
							kr = kg;
							if(kdif < 0) kq = kg;
						}					// endif
					}						// line 1330 (continue)
					aq = agj[kq];
					ar = agj[kr];

					dfacts = (double)(aq*(aq+2) - mg[kq][i]*mg[kq][j])
						            *(ar*(ar+2) - mg[kr][i]*mg[kr][j]);
					dfacts = 0.125 * sqrt(dfacts);

					hh = hh + dfacts*(cupind[kq][kr]-cupdip[kq][kr]);

					if(dodifs == TRUE)
					{
						iv = ivcup[kq][kr];
						if(iv > 0)
						{
							pd[iv] = dfacts;
						}
						iv = ivdip[kq][kr];
						if(iv > 0) pd[iv] = dfacts;
					}

					if((pairg[kq] != 2) || (pairg[kr] != 2))
						goto l1900;

					if((aming[kq] != 2) || (aming[kr] != 2))
						goto l1900;

					bfact = bfactg[aq/2][kq] * bfactg[ar/2][kr];
					cuptmp = cupind[kr][kq] - cupdip[kr][kq];
					fcttmp = dfacts * bfact;
					hh = hh + fcttmp * cuptmp;

					if(dodifs == TRUE)
					{
						iv = ivcup[kr][kq];
						if(iv > 0)
						{
							pd[iv] = fcttmp;
						}
						iv = ivdip[kr][kq];
						if(iv > 0) pd[iv] = fcttmp;
					}					
				}							// endif
				else
				{
					kdifa = -2;
					kdifm = 0;
					aa = agp[i];

					for(kg = 1; kg <= ngrp; kg++)	// do 1450
					{
						mm = mg[kg][i] - mg[kg][j];
						if(abs(mm) > 2) goto l1900;
						ai = aa/agpb[kg];

						if(ai == agj[kg])
						{
							if(mm != 0)
							{
								if((pairg[kg] != 2) || (aming[kg] == 0))
									goto l1900;
								kdifm = kdifm +1;
								ar = ai;
								kr = kg;
							}
						}
						else
						{
							if(abs(ai - agj[kg]) > 2) goto l1900;
							kdifa = kdifa +1;
							if(kdifa > 0) goto l1900;
							if(kdifa < 0)
							{
								kq = kg;
								aq = ai;
								mq = mg[kg][i];
								mmq = mm;
							}
							else
							{
								kr = kg;
								ar = ai;
								mr = mg[kg][i];
								mmr = mm;
							}
						}
						aa = aa - ai * agpb[kg];
					}								// line 1450 (continue)

					if((kdifa + kdifm) > 0) goto l1900;

					sgnval = 0.50;

					if(aq <= agj[kq])
					{
						aq = agj[kq];
						mq = mq - mmq;
						mmq = (-mmq);
					}
						
					if(mmq == 0)
					{
						dfactq = (aq*aq - mq*mq);
						sgnval = 1.0;
					}
					else
					{
						if(mmq > 0)
						{
							sgnval = (-sgnval);
							mq = (-mq);
						}
						dfactq = (aq-mq) * (aq-mq-2);
					}

					dfactq = dfactq*bfactf[aq/2][kq];

					if(kdifa >= 0)
					{
						if(ar <= agj[kr])
						{
							ar = agj[kr];
							mr = mr - mmr;
							mmr = (-mmr);
						}
						
						if(mmr == 0)
							dfactr = (ar*ar - mr*mr);
						else
						{
							if(mmr > 0)
							{
								sgnval = (-sgnval);
								mr = (-mr);
							}
							dfactr = (ar-mr) * (ar-mr-2);
						}

						dfactr = dfactr * bfactf[ar/2][kr];
						cuptmp = cupind[kr][kq] - cupdip[kr][kq];
						fcttmp = sgnval * sqrt(dfactq * dfactr);
						hh = fcttmp * cuptmp;
						
						if(dodifs == TRUE)
						{
							iv = ivcup[kr][kq];
							if(iv > 0)
							{
								pd[iv] = fcttmp;
							}
							iv = ivdip[kr][kq];
							if(iv > 0) pd[iv] = fcttmp;
						}
					}
					else
					{
						if(kdifm != 0)
						{
							dfactr = ar * (ar+2) - mg[kr][i] * mg[kr][j];
							dfacts = sqrt(0.25 * dfactr * dfactq)
								   * bfactg[ar/2][kr];

							if(kr <= kq)
							{
								kg = kr;
								kr = kq;
								kq = kg;
							}

							cuptmp = cupind[kr][kq] - cupdip[kr][kq];
							fcttmp = sgnval * dfacts;
							hh = fcttmp * cuptmp;
							
							if(dodifs == TRUE)
							{
								iv = ivcup[kr][kq];
								if(iv > 0)
								{
									pd[iv] = fcttmp;
								}
								iv = ivdip[kr][kq];
								if(iv > 0) pd[iv] = fcttmp;
							}

						}
						else
						{
							dfactq = sqrt(dfactq);

							for(kr = 1; kr <= ngrp; kr++)	// do 1750
							{
								if(!(pairg[kr] != 2 ||
									 aming[kr] == 0 ||
									 kq == kr))
								{
									if(kq-kr < 0)
									{
										k1 = kr;
										k2 = kq;
									}
									else
									{
										k1 = kq;
										k2 = kr;
									}

									cuptmp = cupind[k1][k2]
									   - cupdip[k1][k2];
									mr = mg[kr][j];
									if(mr != 0)
									{
										ar = agj[kr];
										dfactr = (double)mr;
										dfactr = 0.5 * dfactr
										         * bfactg[ar/2][kr];
										fcttmp = dfactq * dfactr;
										hh = hh + fcttmp * cuptmp;
										if(dodifs == TRUE)
										{
											iv = ivcup[k1][k2];
											if(iv > 0)
											{
												pd[iv] = fcttmp;
											}
											iv = ivdip[k1][k2];
											if(iv > 0) pd[iv] = fcttmp;
										}	// endif
									}		// endif									
								}			// endif
							}				// line 1750 (continue)
						}					// endif (else)
					}						// endif (else)
				}							// endif (else)
			}								// endif
			else	// diagonal element
			{
				for(kr = 1; kr <= ngrp; kr++)	// do 1850
				{
					mr = mg[kr][j];
					ar = agj[kr];

					if(ar != 0)
					{
						for(kq=1; kq <=kr; kq++)
						{
							if(kq == kr)
							{
								dfacts = (double)mr;
								fcttmp = 0.5 * dfacts;
								hh = hh + fcttmp * shift[kr];	// shift contrib.
								if(dodifs == TRUE)
								{
									iv = ivcs[kr];
									if(iv > 0)
									{
										pd[iv] = fcttmp;
									}
								}					// endif
						
								if(pairg[kr] != 2)
								{
									fcttmp = 0.375 * dfacts * dfacts;
									hh = hh + fcttmp * cupdip[kr][kr];
									if(dodifs == TRUE)
									{
										iv = ivdip[kr][kr];
										if(iv > 0) pd[iv] = fcttmp;
									}				// endif
								}					// endif

								if(pairg[kr] == 2)
								{
									dfacts = ar * (ar+2);
									cuptmp = cupind[kr][kr]
										   + 2.0 * cupdip[kr][kr];
									fcttmp = 0.125 * dfacts;
									hh = hh + fcttmp * cuptmp;
									if(dodifs == TRUE)
									{
										iv = ivcup[kr][kr];
										if(iv > 0)
										{
											pd[iv] = fcttmp;
										}
										iv = ivdip[kr][kr];
										if(iv > 0) pd[iv] = 2.0 * fcttmp;
									}				// endif
								}					// endif
							}						// endif
							else					// kq != kr
							{
								aq = agj[kq];
								mq = mg[kq][j];

								if(mq != 0)
								{
									dfacts = 0.25 * (double)(mq * mr);
									cuptmp = cupind[kq][kr]
										   + 2.0 * cupdip[kq][kr];
									hh = hh + dfacts * cuptmp;

									if(dodifs == TRUE)
									{
										iv = ivcup[kq][kr];
										if(iv > 0)
										{
											pd[iv] = dfacts;
										}
										iv = ivdip[kq][kr];
										if(iv > 0) pd[iv] = 2.0 * dfacts;
									}			// endif

									if((pairg[kq] == 2) && (pairg[kr] == 2))
									{
										if((aming[kq] != 0) && (aming[kr] != 0))
										{
											bfact = bfactg[aq/2][kq]
												  * bfactg[ar/2][kr];
											cuptmp = cupind[kr][kq]
												   + 2.0 * cupdip[kr][kq];
											fcttmp = dfacts * bfact;
											hh = hh + fcttmp * cuptmp;

											if(dodifs == TRUE)
											{
												iv = ivcup[kr][kq];
												if(iv > 0)
												{
													pd[iv] = fcttmp;
												}
												iv = ivdip[kr][kq];
												if(iv > 0) pd[iv] = 2.0 * fcttmp;
											}	// endif
										}		// endif
									}			// endif
								}				// endif
							}					// endif (else kq!=kr)
						}						// line 1840 (cont.)
					}		// ar != 0			// endif
				}								// line 1850 (cont.)
			}									// endif (else)
			if(dodifs == TRUE)
			{
				nw = nw + 1;

				if(ne4 <= MPDINT)
				{
					iiint[nw] = ii;
					jjint[nw] = jj;
					for(jv = 1; jv <= nvar; jv++)	// do 1880
						pdint[jv][nw] = pd[jv];
				}
				else
				{
					scratch4 = fopen("scratch.4", "wb");	// IDIF
					fwrite(&ii, sizeof(int), 1, scratch4);
					fwrite(&jj, sizeof(int), 1, scratch4);
					for(int i = 1; i <= nvar; i++)
						fwrite(&pd[i], sizeof(double), 1, scratch4);
					fclose(scratch4);
				}								// endif (else)
			}									// endif

l1900:; // continue
			dh[ii][jj] = hh;
			dh[jj][ii] = hh;	
		}										// line 1950 (continue)
	}											// line 2000 (continue)

	
	// now diagonalize the Hamiltonian

	if(debug)matdump(dh, 1, ne4, 1, ne4, "dh (Hamiltonian)");		// dump Hamiltonian for de bugging
	error = householder(MAXE4, ne4, dh, dj, dk, sortit, sortup);

	if(error == ERROR)
	{
			fprintf(ofp, "\nSIMULATION: hamil(): diagonalization failed!\n");
			fprintf(ofp, "  M(2,3,4), level =   %d   %d   %d\n\n",
				            km2, km32, km43);
			if(debug)fclose(quant);
			if(debug)fclose(energ);
			fflush(ofp);
			return error;
	}

	if(dodifs == TRUE)
	{
		for(iv = 1; iv <= nvar; iv++)							// do 2001
		{
			for(i=1; i<= ne4; i++)								// do 2002
			{
				for(j=1; j <= ne4; j++)
				{
					da[i][j] = 0.0;								// use matclear?
				}												// 2002 continue
			}													// 2002 continue

			if(ne4 <= MPDINT)
			{
				for(i=1; i <= nw; i++)							// do 2003
				{
					ii = iiint[i];
					jj = jjint[i];
					da[ii][jj] = pdint[iv][i];
					da[jj][ii] = pdint[iv][i];
				}												// 2003
			}
			else
			{
//				rewind(scratch4);								// IDIF
				scratch4 = fopen("scratch.4", "rb");			// IDIF
				for(i=1; i<= nw; i++)							// do 2004
				{
					fread(&ii, sizeof(int), 1, scratch4);		// reading IDIF
					fread(&jj, sizeof(int), 1, scratch4);
					for(jv = 1; jv <= nvar; jv++)
					{
						fread(&pd[jv], sizeof(double), 1, scratch4);
					}
					da[ii][jj] = pd[iv];
					da[jj][ii] = pd[iv];
				}												// 2004
				fclose(scratch4);
			}													// endif (else)

			for(i = 1; i <= ne4; i++)							// do 2006
			{
				for(j = 1; j <= ne4; j++)
				{
				db[i][j] = 0.0;
				for(k = 1; k <= ne4; k++)
					db[i][j] = db[i][j] + da[i][k] * dh[k][j];
				}
			}													// 2006

			for(i=1; i <= ne4; i++)								// do 2009
			{
				d[i] = 0.0;
				for(k = 1; k <= ne4; k++)
					d[i] = d[i] + dh[k][i] * db[k][i];
			}													// 2009

			fwrite(&m,  sizeof(int), 1, scratch3);
			fwrite(&n,  sizeof(int), 1, scratch3);
			fwrite(&iv, sizeof(int), 1, scratch3);

			for(i=1; i <= ne4; i++)
			{
				fwrite(&d[i], sizeof(double), 1, scratch3);
			}
		}														// 2001 continue
	}															// endif

	j = 0;

	for(i = m; i <= n; i++)										// do 2010
	{
		j++;
		e3[i] = dj[j];
	}															// 2010 

	for(j = 1; j <= ne4; j++)									// do 2030
	{
		for(i = 1; i <= ne4; i++)
		{
			nu3++;
			u3[nu3] = dh[i][j];
		}
	}															// 2030

	if(prtkey && !iting  && debug)  // write out energy levels as an aid to analysis
	{
//		fprintf(energ, "\n  hamil(): energy levels:\n\n");
		
		l = miso43[km43];

		int temp_spin[20];
		for(i = 1; i <= niso; i++)								// do 2050
		{
			k = l/misopb[i];
			temp_spin[i] = k - misopa[i];
			l = l - k*misopb[i];
		}														// 2050
		
		fprintf(energ, "   mat(4)=%6d  ", km43);
		for(i = 1; i <= niso; i++)
			fprintf(energ, "  %d", temp_spin[i]);
		fprintf(energ, "\n");
		for(i = m; i <= n; i++)
			fprintf(energ, "                %f", e3[i]);
		fprintf(energ, "\n");
	}															// endif

	km43++;

	if(debug)
	{
		fprintf(ofp, "***  hamil(): end of mat(3) km43 = %d   nm43 = %d\n", km43, nm43);
		fflush(ofp);
	}
	if(km43 <= nm43)  goto l1200;  // end of a mat(3)			// CHANGE!!!

	if(debug)fprintf(quant, "\n***  hamil(): dumping to eu:\n");
	int ow = 0;

	ow = ow + fwrite(&nu3, sizeof(int), 1, scratch2);

	if(debug)fprintf(quant, "   ne3 = %d,   ne4 = %d,   nu3 = %d\n", ne3, ne4, nu3);

	for(i=1; i <= ne3; i++)
	{
		ow = ow + fwrite(&e3[i], sizeof(double), 1, scratch2);
		if(debug)fprintf(quant, " %d %f\n", i, e3[i]);
	}
	for(i=1; i <= nu3; i++)
	{
		ow = ow + fwrite(&u3[i], sizeof(double), 1, scratch2);
		if(debug)fprintf(quant, " %d %f\n", i, u3[i]);
	}
	u3_size_local = u3_size_local + nu3;
	if(debug)fprintf(quant, "  objects written = %d\n", ow);
	if(debug)fprintf(quant, "\n");

	if(km32 < nm32) goto l1100;		// end of a mat(2)

	if(km2 < nm2) goto l1000;		// end of last mat(2)

	fclose(scratch1);	// QUANT
	fclose(scratch2);	// EU

	if(dodifs)
	{
		fclose(scratch3);
//		fclose(scratch4);
	}

#ifdef DUMP_LOG
	fp = fopen("matdump.txt", "a");
	fprintf(fp, "\n*** hamil():wrkcom\n");
	fclose(fp);
#endif
	free_vector(shift, 1, MAXGRP);
	free_matrix(cupind, 1, MAXGRP, 1, MAXGRP);
	free_matrix(cupdip, 1, MAXGRP, 1, MAXGRP);
	free_ivector(agj, 1, MAXGRP);				// possible problem here??

	free_vector(pd, 1, MAXVAR);
	free_matrix(pdint, 1, MAXVAR, 1, MPDTRI);
	free_ivector(iiint, 1, MPDTRI);
	free_ivector(jjint, 1, MPDTRI);

	if(debug)fclose(energ);
	if(debug)fclose(quant);

	if(debug)
	{
		fprintf(ofp, "\n***  hamil(): exiting, return code = %d\n", error);
		fprintf(ofp, "***  hamil(): u3 size = %d words\n", u3_size_local);
	}
	fclose(ofp);
	return error;
}

#ifdef DUMP_LOG
#undef DUMP_LOG
#endif
////////////////////////////////////////////////////////////////////////////////


