// Initialize parameters used for iterative simulation

#include <stdio.h>
#include <stdlib.h>
#include "nummrit.h"
#include "maxvalue.h"
#include "matutil.h"
#include <math.h>


// /grpcom/
extern double	*cshift, **cup, **symcup,
				**dip, **symdip;
extern int		*grpiso, *nucg, *pairg,
				*grpspn, fmaxg[MAXGRP+1][MAXGRP+1], *agpb;
extern short	anydip, symspn;
extern char		grpnam[MXGNAM + 1][MAXGRP + 1];

// /spncom/
extern int		ngrp, niso, nvar;

// /prmcom/
extern int		prmsiz, prmnum, prmi, 
				prmj, prmd, lowvar;
extern double	rngdef, rngnit, rngmin;
extern double	*asnhi, *asnlow,
				*rhics,	*rlocs, rngfra,
				**rhiccj, **rloccj,
				**rhiscj, **rloscj,
				**rhiccd, **rloccd,
				**rhiscd, **rloscd;
extern int		*asnmin, *asnnum,
				**prmmap, *ivcs,
				**ivcup, **ivdip;
extern short	chkasn, difzer[];

// /optcom/
extern double	rmscon, rmsdiv, rmslim, damp, asnlim,
				ignlim, itrfac, linwid, asnrms;
extern int		itrlim, itrmin, asnmod;
extern short	anyasn, wgtrms;

void nitprm()
{
	const double smlnum = 0.001,	widfac = 0.75;
	double cupwid, tmpwid;
	int i, j, spinj, tmpmin;


	dclear(MAXGRP, asnlow, 0.0);
	dclear(MAXGRP, asnhi, 0.0);
	iclear(MAXGRP, asnnum, 0);
	iclear(MAXGRP, asnmin, 0);

	for(i = 1; i<=MAXVAR; i++) difzer[i] = TRUE;

	for(i = 1; i <= ngrp; i++)
	{
		asnlow[i] = cshift[i];
		asnhi[i] = cshift[i];
	}

	for(i = 1; i <= ngrp; i++)
	{
		cupwid = 0.0;
		tmpmin = 1;

		if(pairg[i] != 2)	// no two-fold symmetry
		{
			for(j = 1; j <= ngrp; j++)
			{
				if(i != j)
				{
					spinj = abs(nucg[j] * grpspn[j] * pairg[j]);
					if(i < j)
					{
						tmpwid = (fabs(cup[i][j]) + fabs(dip[i][j]));
					}
					else
					{
						tmpwid = (fabs(cup[j][i]) + fabs(dip[j][i]));
					}
					if(tmpwid >= linwid) tmpmin = tmpmin * (spinj+1);
					cupwid = cupwid + tmpwid*spinj;
				}
			}	// 40
		}
		else		// does have two-fold symmetry
		{
			for(j = 1; j <= ngrp; j++)
			{
				spinj = abs(nucg[j] * grpspn[j]);
				if(i == j || pairg[j] == 1)
				{
					if(i < j)
					{
						tmpwid = (fabs(cup[i][j]) + fabs(dip[i][j]));
					}
					else
					{
						tmpwid = (fabs(cup[j][i]) + fabs(dip[j][i]));
					}
					if(tmpwid >= linwid) tmpmin = tmpmin * (spinj+1);
					cupwid = cupwid + tmpwid * spinj;
				}
				else
				{
					tmpwid = (fabs(cup[i][j]) + fabs(dip[i][j]));
					if(tmpwid >= linwid) tmpmin = tmpmin * spinj;

					cupwid = cupwid + tmpwid + tmpwid * spinj;
					tmpwid = (fabs(cup[j][i]) + fabs(dip[j][i]));

					if(tmpwid >= linwid) tmpmin = tmpmin * (spinj + 1);

					cupwid = cupwid + tmpwid*spinj;
				}
			}	// 50
		}
	    if(asnmin[i] < 1) asnmin[i] = 1;
		if(asnmin[i] > 16) asnmin[i] = 16;
		asnlow[i] = asnlow[i] - widfac * cupwid;
		asnhi[i] = asnhi[i] + widfac * cupwid;
	}	// 30

	for(i = 1; i <= ngrp; i++)
	{
		if(rhics[i] > rngdef) rhics[i] = (1.0 + rngfra) * asnhi[i];
		if(rlocs[i] < (-rngdef)) rlocs[i] = (1.0 - rngfra) * asnlow[i];

		for(j = 1; j <= ngrp; j++)
		{
			if(rhiccd[i][j] > rngdef) 
			{
				if(fabs(dip[i][j]) < smlnum)
					rhiccd[i][j] = 2.0;
				else
					rhiccd[i][j] = (1.0 + rngfra) * dip[i][j];
			}

			if(rloccd[i][j] < (-rngdef)) 
			{
				if(fabs(dip[i][j]) < smlnum)
					rloccd[i][j] = -2.0;
				else
					rloccd[i][j] = (1.0 - rngfra) * dip[i][j];
			}

			if(rhiccj[i][j] > rngdef) 
			{
				if(fabs(cup[i][j]) < smlnum)
					rhiccj[i][j] = 0.1;
				else
					rhiccj[i][j] = (1.0 + rngfra) * cup[i][j];
			}

			if(rloccj[i][j] < (-rngdef)) 
			{
				if(fabs(cup[i][j]) < smlnum)
					rloccj[i][j] = -0.1;
				else
					rloccj[i][j] = (1.0 - rngfra) * cup[i][j];
			}
		}
	}

	symto(ngrp, pairg, rhiccj, rhiscj);
	symto(ngrp, pairg, rloccj, rloscj);
	symto(ngrp, pairg, rhiccd, rhiscd);
	symto(ngrp, pairg, rloccd, rloscd);

}

