//  Input simulation parameters and convert to NUMARIT internal parameters

#include <stdio.h>
#include "nummrit.h"
#include "maxvalue.h"
#include "sort_util.h"
#include "matutil.h"
#include "input_sim.h"
#include <string.h>
#include <ctype.h>
#include <math.h>


// /spncom/
extern int ngrp, niso, nvar;

// /optcom/
extern double	rmscon, rmsdiv, rmslim, damp, asnlim,
				ignlim, itrfac, linwid, asnrms;
extern int		itrlim, itrmin, asnmod;
extern short	anyasn, wgtrms;

// /isocom/
extern double	*tiiso, *tliso,
				*fliso, *fuiso;
extern int		*isospn, *ntiso,
				*misopa, *misopb, km2p;
extern char		isonam[MXINAM + 1][MAXISO + 1];
extern char		prtiso[MAXISO + 2];

// /grpcom/
extern double	*cshift, **cup, **symcup,
				**dip, **symdip;
extern int		*grpiso, *nucg, *pairg,
				*grpspn, fmaxg[MAXGRP+1][MAXGRP+1], *agpb;
extern short	anydip, symspn;
extern char		grpnam[MXGNAM + 1][MAXGRP + 1];

// /spncom/
extern int		ngrp, niso, nvar;

// /flagcm/
extern short	autasn, autdel, optmiz, prtptl,
				prtkey, prtdif, prtvcm, pracnd;

// /obscom/
extern int		obscnt, *obslin;
extern double	*obsfrq, *obsint;

// /asncom/
extern double	*asnfrq, *asnwgt, *asndif, *asnclc, asnint;
extern int		asncnt, *asnlin;

// /prmcom/
extern int		*asnmin, *asnnum,
				**prmmap, *ivcs,
				**ivcup, **ivdip;
extern int		prmsiz, prmnum, prmi, 
				prmj, prmd, lowvar;
extern double	*asnhi, *asnlow,
				*rhics, *rlocs, rngfra,
				**rhiccj, **rloccj,
				**rhiscj, **rloscj,
				**rhiccd, **rloccd,
				**rhiscd, **rloscd;

//	DNMR3 common blocks: can probably be removed, is DNMR stuff removed from input routines.
//	common block "permt", permutation vectors and flag for mutual exchange
int		ie[4][6], mu, nconfig;

//	common block "dynpar" 
double	rc[4][4], t2, pop[4];

#define OPTIONS		1
#define TITLE 		2
#define	SHIFTS	    3	
#define	COUPLINGS	4
#define PARAMETERS	5


// external functions
extern void setlwd();

extern int load_options(const char *sim_out);
extern int load_shifts(const char *sim_out);
extern int load_couplings(const char *sim_out);

extern void nitprm();
extern void write_spins(const char *);

// structure holding the spin parameters

int active_system = 1;
struct shift_info shifts[MAX_SHIFTS][MAX_SYSTEMS];
double j_coupling[MAX_SHIFTS][MAX_SHIFTS][MAX_SYSTEMS];
short j_iterate[MAX_SHIFTS][MAX_SHIFTS][MAX_SYSTEMS];
double d_coupling[MAX_SHIFTS][MAX_SHIFTS][MAX_SYSTEMS];
short d_iterate[MAX_SHIFTS][MAX_SHIFTS][MAX_SYSTEMS];

// the structure holding simulated transition information
struct transitions peaks[PEAKMAX];
double unasfreq[MAXASN];
struct transitions asn_trans[MAXASN];

// option strings
char option_label[20][16]={"optimize","autoassign","autoignore",
"linewidth","iterations","transitions","","","",""};

// option values
char option[20][16]={"no","yes","yes","0.1","20","none","","","",""};

int n_shifts[MAX_SYSTEMS];

int				n_trans;
short			calc_mode;
short			expt_exists;	
//extern			CString plot_title;

char			title[80];
char			instring[INPUT_MAX + 1];	// make local


//extern CString sim_out;

//-------------------------------------------------------------------------------------------------
int input(const char *spin_system, const char *sim_out, const char *trans_file)
{
	int i, j;
	short error = FALSE;
	FILE *ofp;

	error = read_spins(spin_system, sim_out, FULL, TRUE);

	if(error == FALSE)
		error = load_options(sim_out);
	if(error == FALSE)
		error = load_shifts(sim_out);
	if(error == FALSE)
		error = load_couplings(sim_out);

	if(error != FALSE)
	{
		ofp = fopen(sim_out, "a");
		fprintf(ofp, "***   Reading of spectral parameters returns an error   ***\n");
		fclose(ofp);
		return error;
	}

	nitprm();		//  optimization initialization

	if(optmiz == TRUE)
	{
		ofp = fopen(sim_out, "a");
		fprintf(ofp, "\nParamaters to vary:\n");
		fprintf(ofp, "\n  #\t param.\t\t range low\t  range hi\n\n");

		for(i = 1; i<=ngrp; i++)
			if(ivcs[i] != 0)
				fprintf(ofp, "  %d\t v[%d]\t\t %9.3f\t %9.3f\n", 
					ivcs[i], i, rlocs[i], rhics[i]);

		for(i=1; i<=ngrp; i++)
			for(j=1; j<=ngrp; j++)
				if(ivcup[i][j] !=0  && ((j > i) || (pairg[i] == 2 && pairg[j] == 2)))
				{
					fprintf(ofp, "  %d\t j[%d][%d]\t %9.3f\t %9.3f\n", 
						ivcup[i][j], i, j, rloccj[i][j], rhiccj[i][j]);
				}

		for(i=1; i<=ngrp; i++)
			for(j=1; j<=ngrp; j++)
				if(ivdip[i][j] !=0  && (j >= i))
				{
					fprintf(ofp, "  %d\t d[%d][%d]\t %9.3f\t %9.3f\n", 
						ivdip[i][j], i, j, rloccd[i][j], rhiccd[i][j]);
				}


 		fprintf(ofp, "\nParameter map:\n");
		for(i=0; i<=nvar; i++)
			fprintf(ofp, "  %2d  %2d  %2d  %2d \n", 
				prmmap[1][i], prmmap[2][i], prmmap[3][i], prmmap[4][i]);

		fclose(ofp);
	}

	if(error == FALSE)
	{
		symto(ngrp, pairg, cup, symcup);
		symto(ngrp, pairg, dip, symdip);
		setlwd();
	}

	if(optmiz == TRUE  && error == FALSE)
	{
		if(error == FALSE)
			error = load_assigned(sim_out, trans_file);
	}

	if(error != FALSE)
	{
		FILE *ofp = fopen(sim_out, "a");
		fprintf(ofp, "***   Reading of assigned transitions returns an error   ***\n");
		fclose(ofp);
	}

	return error;
}

/////////////////////////////////////////////////////////////////////////////////
//  this function reads a NUMMRIT input file.  Again, it is rather complicated
//  in order to allow for the relatively flexible nature of the NUMMRIT input
//  file. 

int read_spins(const char *spin_file, const char *sim_out, short part, short sim)
{

char instring[INPUT_MAX+1];		/* general temporary input string */
char tempstring[INPUT_MAX+1];
char delim[10]; 				/* input delimiters */
char *token_pointer;			/* token extracted by strtok() */
char token1[20], token2[20], token3[20], token4[20], token5[20], token6[20],
	 species[20];
char index1_label[16], index2_label[16];
int index1, index2;
int i, j, species_spin;
FILE *ifp;						/* input file pointer */
FILE *ofp;
FILE *debug_file;				
short status_flag;				/* current read status */
short debug = FALSE;
int error = FALSE;
int par_number = 0;


    (void)strcpy(delim, "= \t");    /* token delimiters */

	if(debug) debug_file = fopen(sim_out, "a");

	ofp = fopen(sim_out, "a");

	if(ofp == NULL)
	{/*
		CString msg;
		msg.Format("Cannot open %s for writing", sim_out);
		AfxMessageBox(msg, MB_ICONERROR);*/
		printf("Cannot open %s for writing\n", sim_out);
		error = ERROR;
		return error;
	}

    fprintf(ofp,"\n\n***  Reading spin system from: %s  ***\n",spin_file);

    ifp = fopen(spin_file, "r");

	if(ifp == NULL)
    {
       fprintf(ofp, "\n***  Error: cannot open %s  ***\n",spin_file);
	   error = ERROR;
	   if(debug)fclose(debug_file);
	   return error;
    }

	fclose(ofp);

       /* clear existing settings */
    if(debug)
            fprintf(debug_file, "SpinWorks: read_spins(): clearing existing spin system\n");
	for(i=0; i < MAX_SHIFTS; i++)
	{
		strcpy(shifts[i][active_system].type, "");
		shifts[i][active_system].freq = 0.0;
		shifts[i][active_system].spin = 1;
		strcpy(shifts[i][active_system].species, "");
        strcpy(shifts[i][active_system].label, "");
		if(part == FULL)
            shifts[i][active_system].iterate = FALSE;
        for(j=0; j < MAX_SHIFTS; j++)
        {
			j_coupling[i][j][active_system] = 0.0;
			if(part == FULL)
                j_iterate[i][j][active_system] = FALSE;
        }
	}
	if(part == FULL)
	{
		for(i=0; i < N_OPTIONS; i++)
		{
			strcpy(option_label[i], "");
			strcpy(option[i], "") ;
		}
    }

	i=0;
	status_flag = OPTIONS;
	n_shifts[active_system] = 0;
	      
	while(fgets(instring, sizeof(instring), ifp) != NULL)
	{
		// remove trailing newline character from input string 
		instring[strlen(instring) - 1] = '\0'; 
		if(debug)fprintf(debug_file,"SpinWorks: read_spins(): instring = %s\n",
			   instring);

         /* initialize tokens to NULL strings */
         token1[0] = '\0'; token2[0] = '\0'; token3[0] = '\0';
         token4[0] = '\0'; token5[0] = '\0'; token6[0] = '\0';
           
         /* a single blank line separates various sections */
         if(strlen(instring) == 0 && status_flag != SHIFTS)
         {
             status_flag++; /* move to next section */
             i=0;
         }

         /* two blank lines between shifts and couplings */
        if((strlen(instring) == 0) && (strlen(tempstring) == 0) &&
			(status_flag == SHIFTS))
	    {
			status_flag = COUPLINGS;
			i=0;
	    }
        (void)strcpy(tempstring, instring); /* save temporary copy */

        if(status_flag == OPTIONS) /* simulation options */
		{
			if(debug)fprintf(debug_file,"OPTIONS:\n");
			if(instring[0] == '#') instring[0] = ' ';
			if(token_pointer = strtok(instring, delim))
	           (void)strcpy(token1, token_pointer);

            if(token_pointer = strtok(NULL, delim))
                   (void)strcpy(token2, token_pointer);

			if(token_pointer = strtok(NULL, delim))
                   (void)strcpy(token3, token_pointer);

			if(token_pointer = strtok(NULL, delim))
			(void)strcpy(token4, token_pointer);

            if(token_pointer = strtok(NULL, delim))
				(void)strcpy(token5, token_pointer);

            if(token_pointer = strtok(NULL, delim))
				(void)strcpy(token6, token_pointer);

            if(debug)fprintf(debug_file,
                  "SpinWorks: read_spins(): option tokens:\n   %s %s %s %s %s %s\n",
                  token1, token2, token3, token4, token5, token6);

            if(strlen(token1) != 0)
		    {
				strcpy(option_label[i], token1);
				strcpy(option[i], token2);
				if(debug)fprintf(debug_file,"  %d   %s  %s\n",
				   i, option_label[i], option[i]);
				i++;
		    }
		    
            if(strlen(token3) != 0)
            {
				strcpy(option_label[i], token3);
                strcpy(option[i], token4);
				if(debug)fprintf(debug_file,"  %d   %s  %s\n",
				   i, option_label[i], option[i]);
                i++;
            }  

            if(strlen(token5) != 0)
            {
				strcpy(option_label[i], token5);
                strcpy(option[i], token6);
				if(debug)fprintf(debug_file,"  %d   %s  %s\n",
				   i, option_label[i], option[i]);
                i++;
            }

		}


        if(status_flag == TITLE) /* case title */
	    {
			if(debug)fprintf(debug_file,"TITLE:\n");
			(void)strcpy(title,instring);
			if(expt_exists == FALSE)
			{
				//plot_title.Format("%s", title);
			}
		}
           
		if(status_flag == SHIFTS) /* species and shifts */
	    {
			if(debug)fprintf(debug_file,"SHIFTS:\n");

			if(token_pointer = strtok(instring, delim))
                 (void)strcpy(token1, token_pointer);

            if(token_pointer = strtok(NULL, delim))
                 (void)strcpy(token2, token_pointer);

            if(token_pointer = strtok(NULL, delim))
				(void)strcpy(token3, token_pointer);

			if(token_pointer = strtok(NULL, delim))
				(void)strcpy(token4, token_pointer);

            if(token_pointer = strtok(NULL, delim))
				(void)strcpy(token5, token_pointer);

            if(debug)fprintf(debug_file,"     %s  %s  %s  %s  %s\n",
				token1, token2, token3, token4, token5);

            if(strcmp(token1, "species") == 0)  /* a new species */
			{
				(void)strcpy(species, token2);
                if(strcmp(token3, "spin") == 0)
					sscanf(token4, "%d", &species_spin);
                else
				species_spin = 1;      /* default to spin 1/2 */
			}

            else if(isdigit(token1[0]) && strlen(tempstring) != 0)  
		    /* same species, new shift */
			{
				(void)strcpy(shifts[i][active_system].type, token1); 
				(void)sscanf(token2, "%lf", &shifts[i][active_system].freq);
				(void)strcpy(shifts[i][active_system].species, species);
				shifts[i][active_system].spin = species_spin;
				if(strcmp(token3, "name") == 0)
                    (void)strcpy(shifts[i][active_system].label, token4);

                if(debug)fprintf(debug_file,"   %d  %s  %lf  %s  %d  %s \n", i, 
				shifts[i][active_system].type, shifts[i][active_system].freq, shifts[i][active_system].species,
				shifts[i][active_system].spin, shifts[i][active_system].label);
                i++;
				n_shifts[active_system]++;
			}
		}

        if(status_flag == COUPLINGS && strlen(tempstring) != 0)
						   /* coupling  constants */
	    {
			if(debug)fprintf(debug_file,"COUPLINGS:\n");

			if(token_pointer = strtok(instring, delim))
                 (void)strcpy(token1, token_pointer);
			if(token_pointer = strtok(NULL, delim))
                 (void)strcpy(token2, token_pointer);
            if(token_pointer = strtok(NULL, delim))
                 (void)strcpy(token3, token_pointer);
            if(token_pointer = strtok(NULL, delim))
                 (void)strcpy(token4, token_pointer);
            if(token_pointer = strtok(NULL, delim))
                 (void)strcpy(token5, token_pointer);

            if(isdigit(token1[0]))                          /* ## form */
            {
				if(debug)fprintf(debug_file,"    ## form");
                strcpy(index1_label, token1);
                strcpy(index2_label, token1);
                index1_label[1] = '\0';
                sscanf(index1_label, "%d", &index1);
                index2_label[0] = ' ';
                sscanf(index2_label, "%d", &index2);
                if(debug)fprintf(debug_file,"   %d   %d\n", index1, index2);
                sscanf(token2, "%lf", 
					&j_coupling[index1-1][index2-1][active_system]);
			}

            else if(token1[0] == 'j' && isdigit(token1[1])) /* j## form */
			{
				// can't do this yet
			}

			else if(token1[0] == 'j' && token1[1] == '(')   /* j(?,?) form */
			{
				if(debug)fprintf(debug_file,"     j(?,?) form,");
				token1[0] = ' '; token1[1] = ' ';
				token1[strlen(token1) - 1] = ' ';
				if(debug)fprintf(debug_file,"   %s\n",token1);
				if(token_pointer = strtok(token1, ", "))
					(void)strcpy(index1_label, token_pointer);
				if(token_pointer = strtok(NULL, ", "))
					(void)strcpy(index2_label, token_pointer);
                index1 = label_to_index(index1_label);
				index2 = label_to_index(index2_label);
				if(debug)fprintf(debug_file,"      indices = %s     %s     %d     %d\n",
				index1_label, index2_label, index1, index2);
                sscanf(token2,"%lf",&j_coupling[index1][index2][active_system]);
				if(debug)fprintf(debug_file,"      coupling = %s %f \n",token2, 
					j_coupling[index1][index2][active_system]);
			}

			else if(token1[0] == 'd' && token1[1] == '(')   /* d(?,?) form */
			{
				int temp;
				if(debug)fprintf(debug_file,"     d(?,?) form,");
                token1[0] = ' '; token1[1] = ' ';
                token1[strlen(token1) - 1] = ' ';

                if(debug)fprintf(debug_file,"   %s\n",token1);
                if(token_pointer = strtok(token1, ", "))
                       (void)strcpy(index1_label, token_pointer);

                if(token_pointer = strtok(NULL, ", "))
                       (void)strcpy(index2_label, token_pointer);

                index1 = label_to_index(index1_label);
                index2 = label_to_index(index2_label);
				/* swap indices if reversed  */
				if(index1 > index2)
				{
					temp = index1;
					index1 = index2;
					index2 = temp;
				}
                if(debug)fprintf(debug_file,"      %s     %s     %d     %d\n",
						index1_label, index2_label, index1, index2);
                sscanf(token2,"%lf",&d_coupling[index1][index2][active_system]);
				anydip = TRUE;
			} 
		}

		if(status_flag == PARAMETERS && strlen(tempstring) != 0) 
					    /* parameters to vary */
		{
			if(debug)fprintf(debug_file,"PARAMETERS TO VARY:\n");

            if(token_pointer = strtok(instring, delim))
                 (void)strcpy(token1, token_pointer);

            if(token_pointer = strtok(NULL, delim))
                 (void)strcpy(token2, token_pointer);

            if(token_pointer = strtok(NULL, delim))
                 (void)strcpy(token3, token_pointer);

            if(isdigit(token1[0]) && strlen(token1) == 1)/* shift, # form */
			{
				par_number ++;
				if(debug)fprintf(debug_file,"   shift, # form\n");
				sscanf(token1, "%d", &index1);
				shifts[index1-1][active_system].iterate = (short)TRUE;
				if(sim)								// running a simulation
				{									// so load parameter maps
					ivcs[index1] = par_number;
					prmmap[prmnum][par_number] = par_number;
					prmmap[prmi][par_number] = index1;
					prmmap[prmj][par_number] = 0;	// since this is a shift
				}
			}
              
			else if(isdigit(token1[0]) && strlen(token1) == 2)
		                                        /* coupling ## form */
			{
				par_number ++;
				if(debug)fprintf(debug_file,"   coupling ## form");
				strcpy(index1_label, token1);
				strcpy(index2_label, token1);
				index1_label[1] = '\0';
				sscanf(index1_label, "%d", &index1);
				index2_label[0] = ' ';
				sscanf(index2_label, "%d", &index2);
				if(debug)fprintf(debug_file,"   &d   %d\n", index1, index2); 
				j_iterate[index1-1][index2-1][active_system] = (short)TRUE;
				if(sim)							// running a simulation, so 
				{								// load parameter map
					ivcup[index1][index2] = par_number;
					prmmap[prmnum][par_number] = par_number;
					prmmap[prmi][par_number] = index1;
					prmmap[prmj][par_number] = index2;	
				}
			}

           else if(token1[0] == 'v' && token1[1] == '(')
							/* shift v(?) form */
			{
				par_number ++;
				if(debug)fprintf(debug_file,"   shift v(?) form\n");
				token1[0] = ' '; token1[1] = ' '; 
				token1[strlen(token1) -1] = ' ';
                if(token_pointer = strtok(token1, ", "))
					(void)strcpy(index1_label, token_pointer);
                index1 = label_to_index(index1_label);
				shifts[index1][active_system].iterate = (short)TRUE;
				if(sim)
				{
					ivcs[index1 + 1] = par_number;
					prmmap[prmnum][par_number] = par_number;
					prmmap[prmi][par_number] = index1 + 1;
					prmmap[prmj][par_number] = 0;	// since this is a shift
				}
			}

			else if(token1[0] == 'j' && token1[1] == '(')
				        	/* coupling j(?,?) form */
			{
				par_number ++;
				if(debug)fprintf(debug_file,"   coupling j(?,?) form\n");
				token1[0] = ' '; token1[1] = ' ';
				token1[strlen(token1) -1] = ' ';
				if(token_pointer = strtok(token1, ", "))
					(void)strcpy(index1_label, token_pointer);
                index1 = label_to_index(index1_label);
				if(token_pointer = strtok(NULL, ", "))
					(void)strcpy(index2_label, token_pointer);
                index2 = label_to_index(index2_label);	
				j_iterate[index1][index2][active_system] = (short)TRUE;
				if(sim)
				{
					ivcup[index1+1][index2+1] = par_number;
					prmmap[prmnum][par_number] = par_number;
					prmmap[prmi][par_number] = index1 + 1;
					prmmap[prmj][par_number] = index2 + 1;	
				}
			}

			else if(token1[0] == 'd' && token1[1] == '(')
                                                /* coupling d(?,?) form */
			{
				par_number ++;
				int temp;
                if(debug)fprintf(debug_file,"   coupling d(?,?) form\n");
                token1[0] = ' '; token1[1] = ' ';
                token1[strlen(token1) -1] = ' ';
                if(token_pointer = strtok(token1, ", "))
					(void)strcpy(index1_label, token_pointer);
                index1 = label_to_index(index1_label);
                if(token_pointer = strtok(NULL, ", "))
                    (void)strcpy(index2_label, token_pointer);
                index2 = label_to_index(index2_label);
                /* swap indices if reversed */
                if(index1 > index2)
                {
					temp = index1;
                    index1 = index2;
                    index2 = temp;
                }
                d_iterate[index1][index2][active_system] = (short)TRUE;
				if(sim)
				{
					ivdip[index1+1][index2+1] = par_number;
					prmmap[prmnum][par_number] = par_number;
					prmmap[prmi][par_number] = index1 + 1;
					prmmap[prmj][par_number] = index2 + 1;
					// need more here ???!!!
				}
			}
		}
	} /* end while */

	if(calc_mode == DNMR || calc_mode == HIRES)	// check for any DNMR parameters appended to
	{							// end of spin system file (JCAMP DX Format)
		char param[80];
		rewind(ifp);				

		while(fgets(instring, sizeof(instring), ifp) != NULL)
		{
			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$RC[1][2]="))
				sscanf(instring, "%s %lf", param, &rc[1][2]);

			sscanf(instring, "%s" , param);
			if(!strcmp(param,"##$RC[1][3]="))
				sscanf(instring, "%s %lf", param, &rc[1][3]);

			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$RC[2][3]="))
				sscanf(instring, "%s %lf", param, &rc[2][3]);

			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$T2="))
				sscanf(instring, "%s %lf", param, &t2);

			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$MU="))
				sscanf(instring, "%s %d", param, &mu);

			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$NCONFIG="))
				sscanf(instring, "%s %d", param, &nconfig);

			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$PVEC2="))
				sscanf(instring, "%s %1d %1d %1d %1d %1d", param, 
					&ie[2][1], &ie[2][2], &ie[2][3], &ie[2][4], &ie[2][5]);

			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$PVEC3="))
				sscanf(instring, "%s %1d %1d %1d %1d %1d", param, 
					&ie[3][1], &ie[3][2], &ie[3][3], &ie[3][4], &ie[3][5]);

			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$POP[1]="))
				sscanf(instring, "%s %lf", param, &pop[1]); 

			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$POP[2]="))
				sscanf(instring, "%s %lf", param, &pop[2]);

			sscanf(instring, "%s", param);
			if(!strcmp(param,"##$POP[3]="))
				sscanf(instring, "%s %lf", param, &pop[3]);
		}
	}

    /* close file and make this the current  spin system */
	if(debug)fclose(debug_file);
    fclose(ifp);

 //	write_spins(spin_system);	//************
            
	nvar = par_number;
	return error;
}


/*****************************************************************************/
/* this function turns takes a label string and returns the corresponding
   index number */

int label_to_index(char *label_string)
   
{
int i, index;

for(i=0; i<MAX_SHIFTS; i++)
   {
   if(strcmp(label_string, shifts[i][active_system].label) == 0)
      {
      index = i;
      break;
      }
   }

return index;
}

////////////////////////////////////////////////////////////////////////////////
// read assigned transitions

int get_assigned(const char *sim_out, const char *trans_file, int *n_unasfreq)
{
	int i, k;
	FILE *ifp;
	FILE *ofp;
	short debug = FALSE;				// FALSE for production
	int n_values;
	double a;
	int n;

	ifp = fopen(trans_file,"r");
	ofp = fopen(sim_out, "a");
	fprintf(ofp, "\nReading assigned transitions\n\n");
    if(ifp==NULL)
    {
		fprintf(ofp,
			"*** Error: get_trans(): cannot open assigned transition file: %s\n",
                    trans_file);
		fclose(ofp);
		return ERROR;
    }

    i=0; k=0;
    while((getline(sizeof(instring), ifp) > 0) && (i < MAXASN) && (k < MAXASN))
	{
		if(debug)
			fprintf(ofp,"***  get_assigned(): input string= %s",instring);

         n_values=sscanf(instring,"%lf line=%d",&a,&n);
		if(debug)
			fprintf(ofp,"***  get_assigned(): nvalues= %d  a= %f  n= %d\n\n",n_values,a,n);

		if(n_values==2)
	    {
			asn_trans[k].trans_no = n;
			asn_trans[k].obs_freq = a;
			k++;
		}

		else if(n_values==1)
	    {
			unasfreq[i]=a;
			i++;
	    }

	}  // end while 

// print summary of assigned transitions found 

	fprintf(ofp,"   %d assigned transitions read \n",k);
	if(i != 0)
        fprintf(ofp,"   %d frequencies with no assignments read \n",i);

	(void)fclose(ofp);
    (void)fclose(ifp);	// close the assigned transition file

	*n_unasfreq = i;	// return number of observed but unassigned frequencies
	return k;			// return number of transitions read	
}

///////////////////////////////////////////////////////////////////////////

int getline(int lim, FILE *ifp)

/*  get a line from the stream pointed to by the input file pointer
    ifp and put the string in instring.  The string may have a maximum
    length of lim and the return value is the number of 
    characters actually read.
*/
{
	int c,i;
	for(i=0; i<lim-1 && (c=getc(ifp))!=EOF && c!='\n'; i++)
		instring[i]=c;
	if(c=='\n')
	{
		instring[i]=c;
		i++;
	}	
	instring[i]='\0';
	return i;
}

///////////////////////////////////////////////////////////////////////////

int get_string(int lim, char *str,  FILE *ifp)

/*  get a line from the stream pointed to by the input file pointer
    ifp and put the string in instring.  The string may have a maximum
    length of lim and the return value is the number of 
    characters actually read.
*/
{
	int c,i;
	for(i=0; i<lim-1 && (c=getc(ifp))!=EOF && c!='\n'; i++)
		str[i]=c;
	//if(c=='\n')
	//{
	//	instring[i]=c;
	//	i++;
	//}	
	str[i]='\0';
	return i;
}
///////////////////////////////////////////////////////////////////////////////////

int load_assigned(const char *sim_out, const char *trans_file)

//load frequencies and line numbers into asn... and obs... arrays
{
	int error = FALSE;			// FALSE for production
	int assignments_read;		// assignments returned by get_assigned()
	int obs_read;				// number of unassigned frequencies read
	int i;
	double intens = 1.0;		// sometime do actual intensity, kludge for now

	FILE *ofp = fopen(sim_out, "a");
	assignments_read = get_assigned(sim_out, trans_file,  &obs_read);
	obscnt = 0;	asncnt = 0;

	for(i = 0; (i < assignments_read) && (asncnt <= MAXASN) && (obscnt <= MAXOBS)
		&& i <= MAXASN; i++)
	{
		if(fabs(asn_trans[i].obs_freq) > 0.0000001)
		{
			asncnt++;
			obscnt++;
			asnfrq[asncnt] = asn_trans[i].obs_freq;
			asnlin[asncnt] = asn_trans[i].trans_no;
			asnwgt[asncnt] = intens;
			obsfrq[obscnt] = asn_trans[i].obs_freq;
			obsint[obscnt] = intens;
			obslin[obscnt] = asn_trans[i].trans_no;
		}
	}

	for(i = 0; (i < obs_read) && (obscnt <= MAXOBS);  i++)
	{
		if(fabs(unasfreq[i]) > 0.0000001)
		{
			obscnt++;
			obsfrq[obscnt] = unasfreq[i];
			obsint[obscnt] = intens;
		}
	}


	stobsf(MAXOBS, obscnt, obslin, obsfrq, obsint);
	stasnl(MAXASN, asncnt, asnlin, asnfrq, asnwgt, asnclc, asndif);
	killdupe(sim_out, MAXASN, asncnt, asnlin, asnfrq, asnwgt, asnclc, asndif);


	fprintf(ofp, "\n   Observed transitions read =  %8d\n", obscnt);
	fprintf(ofp, "   Total asigned line numbers = %8d\n", asncnt);

	if(asncnt < nvar)
	{
		fprintf(ofp, 
		 "\n***  The number of assigned frequencies is less than the number of variables: %4d\n",
		 nvar);
		if(autasn == TRUE)
			fprintf(ofp, "    an initial auto assignment will be done\n");
		else
			error = ERROR;
	}
	fclose(ofp);
	return error;
}

/////////////////////////////////////////////////////////////////////////

void setlwd()
{
	int i, j;
	const double smlcup = 0.005;
	double mincup = 1000000.0;

	for(i=1; i<=ngrp; i++)
	{
		for(j=1; j<=ngrp; j++)
		{
			if(fabs(cup[i][j]) >= smlcup)	
				 mincup = fmin(mincup, fabs(cup[i][j]));
		}
	}

	if(anydip == TRUE)
	{
		for(i=1; i<=ngrp; i++)
		{
			for(j=1; j<=ngrp; j++)
			{
				if(fabs(dip[i][j]) >= smlcup)
					mincup = fmin(mincup, fabs(dip[i][j]));
			}
		}
	}

    linwid = fmin(linwid, 0.667*mincup);    
}




