// load options from a NUMMRIT input file

#include <stdio.h>
#include <stdlib.h>
#include "nummrit.h"
#include "maxvalue.h"
#include <string.h>
#include <math.h>


// /optcom/
extern double	rmscon, rmsdiv, rmslim, damp, asnlim,
				ignlim, itrfac, linwid, asnrms;
extern int		itrlim, itrmin, asnmod;
extern short	anyasn, wgtrms;

// /prmcom/
extern int		prmsiz, prmnum, prmi, 
				prmj, prmd, lowvar;
extern double	rngdef, rngnit, rngmin;
extern double	*asnhi, *asnlow,
				*rhics, *rlocs, rngfra,
				**rhiccj, **rloccj,
				**rhiscj, **rloscj,
				**rhiccd, **rloccd,
				**rhiscd, **rloscd;
extern int		*asnmin, *asnnum,
				**prmmap, *ivcs, // prmap runs from 0 to 4  CHECK!!!
				**ivcup, **ivdip;
extern short	chkasn, difzer[];

// /spncom/
extern int		ngrp, niso, nvar;

// /flagcm/
extern short	autasn, autdel, optmiz, prtptl,
				prtkey, prtdif, prtvcm, pracnd;

// /sioopt/
extern short	echoit, tracit;

// /shpcom/
extern double	ampscl, shpfac;

extern char option_label[20][16];
extern char option[20][16];

//extern CString sim_out;

/////////////////////////////////////////////////////////////////////////////////////////

int	load_options(const char *sim_out)
{
	// loads the spectral parameters from the Xsim read_spins()
	// function into the variables used by NUMMRIT.
	// A bit messy since we are translating (ughh) Fortran.
	// Also initializes some defaults

	int error = FALSE;
	const double small_val = 0.000001;
	short pltout;
	short debug = FALSE;
	FILE *ofp = fopen(sim_out, "a");

	// initialize prmcom
	chkasn = TRUE;
	rngfra = 0.25;

	// initialize flagcm
	optmiz = FALSE;		prtkey = FALSE;		prtptl = FALSE;
	prtdif = FALSE;		prtvcm = FALSE;		pracnd = FALSE;
	autdel = FALSE;		autasn = FALSE;

	// initialize optcom
	itrlim = 15;		itrmin = 1;			asnmod = 3;
	asnrms = 0.1;		linwid = 0.05;		rmscon = 0.10;
	rmsdiv = 0.10;		rmslim = 0.015;		damp = 1.0;
	anyasn = FALSE;		wgtrms = FALSE;		itrfac = 0.99;
	ignlim = 3.0;		asnlim = 0.90 * ignlim;


	// /shpcom/
	ampscl = 1.0;		shpfac = 1.0;		pltout = TRUE;

	// /sioopt/
	echoit = TRUE;		tracit = TRUE;

	// now scan options read by Xsim style input function and load
	// into nummrit parameters
	for(int i=0; i < N_OPTIONS ; i++)
	{
		if(strcmp("ampscale", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &ampscl);
			ampscl = fabs(ampscl);
			if(ampscl < small_val) ampscl = 1.0;
			if(debug)fprintf(ofp,"ampscale = %f\n",ampscl);
		}	// endif

		else if(strcmp("assignmod", option_label[i]) == 0)
		{
			sscanf(option[i],"%d", &asnmod);
			asnmod = abs(asnmod);
			if(asnmod < 2) asnmod = 2;
			if (asnmod > 10) asnmod = 10;
			if(debug)fprintf(ofp,"asnmod = %d\n",asnmod);
		}	// endif

		else if(strcmp("assignrms", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &asnrms);
			asnrms = fabs(asnrms);
			if(asnrms < small_val) asnrms = 0.10;
			if(asnrms > 10.) asnrms = 10.0;
			if(debug)fprintf(ofp,"assignrms = %f\n",asnrms);
		}	// endif

		else if(strcmp("autoassign", option_label[i]) == 0)
		{
			if(strcmp("yes", option[i]) == 0)
				autasn = TRUE;
			else
				autasn = FALSE;
			if(debug)fprintf(ofp,"autasn = %d\n",autasn);
		}

		else if(strcmp("autoignore", option_label[i]) == 0)
		{
			if(strcmp("yes", option[i]) == 0)
				autdel = TRUE;
			else
				autdel = FALSE;
			if(debug)fprintf(ofp,"autdel = %d\n",autdel);
		}

		else if(strcmp("assignlimit", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &asnlim);
			if(debug)fprintf(ofp,"asnlim = %f\n",asnlim);
		}	// endif

		else if(strcmp("checkassign", option_label[i]) == 0)
		{
			if(strcmp("yes", option[i]) == 0)
				chkasn = TRUE;
			else
				chkasn = FALSE;
			if(debug)fprintf(ofp,"chkasn = %d\n",chkasn);
		}

		else if(strcmp("damping", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &damp);
			damp = fabs(ampscl);
			if(damp < small_val) damp = 1.0;
			if(debug)fprintf(ofp,"damp = %f\n",damp);
		}	// endif

		else if(strcmp("ignorelimit", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &ignlim);
			if(debug)fprintf(ofp,"ignnlim = %f\n",ignlim);
		}	// endif

		else if(strcmp("iterations", option_label[i]) == 0)
		{
			sscanf(option[i],"%d", &itrlim);
			itrlim = abs(itrlim);
			if(itrlim == 0) itrlim = 25;
			if (asnmod > 99) itrlim = 99;
			if(debug)fprintf(ofp,"itrlim = %d\n",itrlim);
		}	// endif

		else if(strcmp("iterfactor", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &itrfac);
			if(debug)fprintf(ofp,"itrfac = %f\n",itrfac);
		}	// endif

		else if(strcmp("linewidth", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &linwid);
			linwid = fabs(linwid);
			if(linwid <= small_val) linwid = 0.050;
			if(linwid > 2.0) linwid = 2.0;
			if(debug)fprintf(ofp,"linewidth = %f\n",linwid);
		}	// endif

		else if(strcmp("optimize", option_label[i]) == 0)
		{
			if(strcmp("yes", option[i]) == 0)
				optmiz = TRUE;
			else
				optmiz = FALSE;
			if(debug)fprintf(ofp,"optimize = %d\n",optmiz);
		}

		//  plot and print options eliminated as pretty much useless

		else if(strcmp("rangefraction", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &rngfra);
			asnrms = fabs(rngfra);
			if(rngfra <= small_val) rngfra = rngmin;
			if(debug)fprintf(ofp,"rangefraction = %f\n",rngfra);
		}	// endif

		else if(strcmp("rmsconverge", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &rmscon);
			asnrms = fabs(rmscon);
			if(rmscon <= small_val) rmscon = 0.01;
			if(rmscon > 0.25) rmscon = 0.25;
			if(debug)fprintf(ofp,"rmsconverge = %f\n",rmscon);
		}	// endif

		else if(strcmp("rmsdiverge", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &rmsdiv);
			asnrms = fabs(rmsdiv);
			if(rmsdiv <= small_val) rmsdiv = 0.01;
			if(rmsdiv > 0.25) rmsdiv = 0.25;
			if(debug)fprintf(ofp,"rmsdiverge = %f\n",rmsdiv);
		}	// endif

		else if(strcmp("rmslimit", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &rmslim);
			rmslim = fabs(rmslim);
			if(rmslim <= small_val) rmslim = 0.005;
			if(rmslim > 10.0) rmslim = 10.0;
			if(debug)fprintf(ofp,"rmslimit = %f\n",rmslim);
		}	// endif

		else if(strcmp("routinetrace", option_label[i]) == 0)
		{
			if(strcmp("yes", option[i]) == 0)
				tracit = TRUE;
			else
				tracit = FALSE;
			if(debug)fprintf(ofp,"routinetrace = %d\n",tracit);
		}

		else if(strcmp("shapefactor", option_label[i]) == 0)
		{
			sscanf(option[i],"%lf", &shpfac);
			shpfac = fabs(shpfac);
			if(shpfac <= small_val) shpfac = 1.0;
			if(debug)fprintf(ofp,"shapefactor = %f\n",shpfac);
		}	// endif

		// useless short form question eliminated
		// transition printing stuff eliminated (at least for now)
		
		else if(strcmp("weightrms", option_label[i]) == 0)
		{
			if(strcmp("yes", option[i]) == 0)
				wgtrms = TRUE;
			else
				wgtrms = FALSE;
			if(debug)fprintf(ofp,"weightrms = %d\n",wgtrms);
		}

		else if(strcmp("printdif", option_label[i]) == 0)
		{
			if(strcmp("yes", option[i]) == 0)
				prtdif = TRUE;
			else
				prtdif = FALSE;
			if(debug)fprintf(ofp,"prtdif = %d\n",autasn);
		}

		else
		{
			if(strlen(option_label[i]) > 0)
				fprintf(ofp, "\n***\n*** SIMULATION: unrecognized option: %s\n***\n",
					option_label[i]);
		}
	}

	// do equivalent of subroutine nitaut
	// basically force reasonable parameters
	itrfac = fabs(itrfac);
	if(itrfac > 1.0) itrfac = 1.0;
	if (itrfac < 0.5) itrfac = 0.5;

	asnlim = fabs(asnlim);
	if(asnlim <= 0.001) asnlim = 1.92;
	if(asnlim < 1.28) asnlim = 1.28;

	ignlim = fabs(ignlim);
	if(ignlim <= 0.001) ignlim = 2.4;
	if(ignlim < 1.60) ignlim = 1.60;
	if(echoit == TRUE)		// echo the input
	{
		fprintf(ofp,"\nOptions:\n\n");
		fprintf(ofp, "   Opimize shifts and couplings? =   %7d\n", optmiz);
		fprintf(ofp, "   Ignore bad transitions? =         %7d\n", autdel);
		fprintf(ofp, "   Auto assign observed peaks? =     %7d\n", autasn);
		fprintf(ofp, "   Maximum number of iterations? =   %7d\n", itrlim);
		fprintf(ofp, "   RMS convergence limit =           %7.3f\n", rmslim);
		fprintf(ofp, "   RMS below this for autoassign =   %7.3f\n", asnrms);
		fprintf(ofp, "   Trans. freq. this * RMS ignored = %7.3f\n", ignlim);

		// add more sometime when I feel like "monkey coding!"
	}

	fclose(ofp);
	return error;
}

