// Load the Xsim/SpinWorks type couplings into the NUMARIT parameters

#include <stdio.h>
#include "nummrit.h"
#include "maxvalue.h"
#include <string.h>


//#include "SpinWorks.h"

// /grpcom/
extern double	*cshift, **cup, **symcup,
				**dip, **symdip;
extern int		*grpiso, *nucg, *pairg,
				*grpspn, fmaxg[MAXGRP+1][MAXGRP+1], *agpb;
extern short	anydip, symspn;
extern char		grpnam[MXGNAM + 1][MAXGRP + 1];

// /isocom/
extern double	*tiiso, *tliso,
				*fliso, *fuiso;
extern int		*isospn, *ntiso,
				*misopa, *misopb, km2p;
extern char		isonam[MXINAM + 1][MAXISO + 1];
extern char		prtiso[MAXISO + 2];

// /optcom/
extern double	rmscon, rmsdiv, rmslim, damp, asnlim,
				ignlim, itrfac, linwid, asnrms;
extern int		itrlim, itrmin, asnmod;
extern short	anyasn, wgtrms;

// /prmcom/
extern int		prmsiz, prmnum, prmi, 
				prmj, prmd, lowvar;
extern double	rngdef, rngnit, rngmin;
extern double	*asnhi, *asnlow,
				*rhics, *rlocs, rngfra,
				**rhiccj, **rloccj,
				**rhiscj, **rloscj,
				**rhiccd, **rloccd,
				**rhiscd, **rloscd;
extern int		*asnmin, *asnnum,
				**prmmap, *ivcs, // prmap runs from 0 to 4  CHECK!!!
				**ivcup, **ivdip;
extern short	chkasn, difzer[];

// /spncom/
extern int		ngrp, niso, nvar;

// /flagcm/
extern short	autasn, autdel, optmiz, prtptl,
				prtkey, prtdif, prtvcm, pracnd;

// /sioopt/
extern short	echoit, tracit;

extern struct	shift_info shifts[MAX_SHIFTS][MAX_SYSTEMS];
extern int		n_shifts[MAX_SYSTEMS];
extern int		active_system;

int ksym, nsym;

//extern CString sim_out;

/////////////////////////////////////////////////////////////////////////////////////////

int load_shifts(const char *sim_out)
// load shifts read by Xsim input routine into 
// nummrit parameters
{
	int error = FALSE;
	short symtry = FALSE;
	short debug = FALSE;		// FALSE for production
	int tmpint;
	FILE *ofp = fopen(sim_out, "a");

	niso = 1;

	if(debug)fprintf(ofp, "SIMULATION: load_shifts(): n_shifts = %d\n",n_shifts[active_system]);
	for(ngrp = 1; ngrp <= n_shifts[active_system]; ngrp++)
	{
		if(debug)fprintf(ofp, 
			"SIMULATION: load_shifts(): extracting shifts ngrp = %d\n", ngrp);
		double rnghi = rngnit; double rnglow = -rngnit;

		if(ngrp != 1)
		{
			if(strcmp(shifts[ngrp-2][active_system].species, shifts[ngrp-1][active_system].species) != 0)
				niso ++;
		}

		if(shifts[ngrp-1][active_system].type[1] == '*')
		{
			tmpint = 3;	//  ??
			symtry = TRUE;
			if(debug)fprintf(ofp, "SIMULATION: load_shifts(): symmetry found\n");
		 
			if(shifts[ngrp-1][active_system].type[0] != '2')
			{
				fprintf(ofp, "SIMULATION ERROR: load_shifts():  incorrect symmetry %s\n",
				     shifts[ngrp][active_system].type);
				fprintf(ofp, "         only a value of 2*n is supported\n");
				error = ERROR;
			}
			if(shifts[ngrp-1][active_system].type[2] == '1')
				nucg[ngrp] = 1;
			else if(shifts[ngrp-1][active_system].type[2] == '2')
				nucg[ngrp] = 2;
			else if(shifts[ngrp-1][active_system].type[2] == '3')
				nucg[ngrp] = 3;
		}
		else
		{
			tmpint = 1;
			symtry = FALSE;
			sscanf(shifts[ngrp-1][active_system].type, "%d", &nucg[ngrp]);
		}

		// load chemical shift values;
		cshift[ngrp] = shifts[ngrp-1][active_system].freq;

		// load group name
		// must add dupe checking
		if(strlen(shifts[ngrp-1][active_system].label) == 0)
			sprintf(grpnam[ngrp], "%2d", ngrp);
		else
			(void)strcpy(grpnam[ngrp], shifts[ngrp-1][active_system].label);

		// load spin values
		if(nucg[ngrp] <= 1 || isospn[niso] <= 1)
		{
			if(nucg[ngrp] == 0) nucg[ngrp] = 1;
			grpiso[ngrp] = niso;
			isospn[niso] = shifts[ngrp-1][active_system].spin;
			grpspn[ngrp] = shifts[ngrp-1][active_system].spin;

			if(debug)fprintf(ofp, "ngrp = %d:   grpiso = %d   isospn[%d] = %d   grpspn = %d\n",
				ngrp, grpiso[ngrp], niso, isospn[niso], grpspn[ngrp]);
			fflush(ofp);

			fmaxg[1][ngrp] = grpspn[ngrp] * nucg[ngrp];
			
			if(debug)fprintf(ofp, "***************\n");
			fflush(ofp);

			if(fmaxg[1][ngrp] > MAXF)
			{
				fprintf(ofp, "SIMULATION: ERROR:  total spin limit for group %d exceeded\n",
					ngrp);
				error = ERROR;
			}	// endif

		}	// endif

		rhics[ngrp] = cshift[ngrp] + rnghi;		// maybe make this a bit
		rlocs[ngrp] = cshift[ngrp] + rnglow;	// fancier some time

		if(symtry == TRUE)
		{
			ksym = ngrp;
			nsym = nsym +1;
			fmaxg[2][ngrp] = fmaxg[1][ngrp];
			pairg[ngrp] = 2;
		}
		else
		{
			fmaxg[2][ngrp] = 0;
			pairg[ngrp] = 1;
		}
	}

	if(nsym > 0)								// we have symmetry
	{
		symspn = TRUE;
		if(nsym == 1)
		{
			nsym = 2 * nucg[ksym];
			fprintf(ofp, "/n***  Note: pair %2d is %2d magnetically equivalent nuclei\n",
				ksym, nsym);
		}
	}

	ngrp = ngrp -1;
	if(debug)fprintf(ofp, "SIMULATION: load_shifts(): ngrp = %d\n",ngrp);
	if(echoit == TRUE || tracit == TRUE)
	{
		fprintf(ofp, "\n\nGroups and chemical shifts:\n\n");
		fprintf(ofp, "  #\tname\tshift\t\tspins\tspecies\tspin\tsym \n\n");
		for(int i=1; i<=ngrp; i++)
		{
			fprintf(ofp, " %2d\t %6s\t %9.3f\t %2d\t %2d\t %3d\t %2d \n",
				     i, grpnam[i], cshift[i], nucg[i], grpiso[i], 
					 shifts[i-1][active_system].spin, pairg[i]);
		}
	}

	fclose(ofp);
	return error;
}


