//  matrix diagonalization routines

#include <stdio.h>
#include "nummrit.h"
#include "matrix_diag.h"
#include <math.h>


/////////////////////////////////////////////////////////////////////////////////////////
int householder(int nmax, int nsize, double **evec, double *eval, 
				double *tmp, short sortit, short sortup)

{
	int error = FALSE;

	tred2(nmax, nsize, evec, eval, tmp);

	error = tqli(nmax, nsize, evec, eval, tmp);

	if((error == FALSE)  && (sortit == TRUE))
	{
		eigsort(nmax, nsize, evec, eval, sortup);
	}

	return error;
}

///////////////////////////////////////////////////////////////////////////////

void tred2(int nmax, int nsize, double **a, double *d, double *e)
{
	double f, g, h, hh, scale;
	int i, j, k, l;

	if(nsize > 1)
	{
		for(i=nsize; i >= 2; i--)
		{
			l = i - 1;
			h = 0.0;	scale = 0.0;
			if(l > 1)
			{
				for(k = 1; k <= l; k++)
					scale = scale + fabs(a[i][k]);

				if(scale == 0)
				{
					e[i] = a[i][l];
				}
				else
				{
					for(k = 1; k <= l; k++)
					{
						a[i][k] = a[i][k] / scale;
						h = h + a[i][k] * a[i][k];
					}

					f = a[i][l];
					g = f>0 ? -sqrt(h) : sqrt(h);  // check what this does!!
					e[i] = scale * g;
					h = h - f * g;
					a[i][l] = f - g;
					f = 0.0;

					for(j = 1; j<= l; j++)
					{
						a[j][i] = a[i][j] / h;
						g = 0.0;
						for(k = 1; k <= j; k++)
							g = g + a[j][k] * a[i][k];
						if(l > j)
						{
							for(k = j+1; k <= l; k++)
								g = g + a[k][j] * a[i][k];
						}
						e[j] = g / h;
						f = f + e[j] * a[i][j];
					}
					hh = f / (h + h);
					for(j = 1; j <= l; j++)
					{
						f = a[i][j];
						g = e[j] - hh * f;
						e[j] = g;
						for(k = 1; k <= j; k++)
							a[j][k] = a[j][k] - f*e[k] - g*a[i][k];
					}
				}		// endif
			}
			else
			{
				e[i] = a[i][l];
			}
			d[i] = h;
		}
	}

	d[1] = 0.0;		e[1] = 0.0;

	for(i=1; i <=nsize; i++)
	{
		l = i - 1;

		if(d[i] != 0.0)
		{
			for(j = 1; j <= l; j++)
			{
				g = 0.0;
				for(k = 1; k <= l; k++)
					g = g + a[i][k] * a[k][j];
				for(k = 1; k <= l; k++)
					a[k][j] = a[k][j] - g*a[k][i];
			}
		}
		d[i] = a[i][i];
		a[i][i] = 1.0;

		if(l >= 1)
		{
				for(j = 1; j<= l; j++)
			{
				a[i][j] = 0.0;
				a[j][i] = 0.0;
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////

int tqli(int nmax, int nsize, double **z, double *d, double *e)
{
	int error = FALSE;
	double b, c, dd,  f, g, p, r, s;
	int iter, i, k, l, m;
	const int maxitr = 30;
	FILE *ofp;

	if(nsize <=1 ) return error;

	for(i = 2; i<= nsize; i++)
		e[i-1] = e[i];
	
	e[nsize] = 0.0;

	for(l = 1; l<=nsize; l++)
	{
		iter = 0;

		do
		{
			for(m = l; m <= nsize-1; m++)
			{
				dd = fabs(d[m]) + fabs(d[m+1]);
				if(fabs(e[m]) + dd == dd)break;
			}
			if(m != l)
			{
				if(iter++ == maxitr) 
				{
					/*ofp = fopen(sim_out, "a");
					fprintf(ofp, "SIMULATION: tqli(): iter = %d,  maxitr = %d\n",
						iter, maxitr);
					fclose(ofp);*/
					return ERROR;
				}
				g = (d[l+1] - d[l])/(2.0 * e[l]);
				r = sqrt((g*g) + 1.0);
				g = d[m] - d[l] + e[l]/(g + SIGN(r,g));
				s = 1.0;	c = 1.0;	p = 0.0;

				for(i = m-1; i >= l; i--)
				{
					f = s * e[i];
					b = c * e[i];
					if(fabs(f) >= fabs(g))
					{
						c = g / f;
						r = sqrt(((c*c) + 1.0));
						e[i+1] = f * r;
						c *= (s = 1.0/r);
					}
					else
					{
						s = f/g;
						r = sqrt((s*s) + 1.0);
						e[i+1] = g * r;
						s *= (c = 1.0/r);
					}
					g = d[i+1] - p;
					r = (d[i] - g) * s + 2.0 * c * b;
					p = s * r;
					d[i+1] = g + p;
					g = c * r - b;
					for(k = 1; k <= nsize; k++)
					{
							f = z[k][i+1];
							z[k][i+1] = s * z[k][i] + c * f;
							z[k][i]   = c * z[k][i] - s * f;
					}
				}
				d[l] = d[l] - p;
				e[l] = g;
				e[m] = 0.0;
			}
		}  while (m != l);
	}
	return error;
}

////////////////////////////////////////////////////////////////////////////////

void eigsort(int nmax, int nsize, double **evec, double *eval, short sortup)
{
	int i, j, k;
	double tmp;

	if(sortup == TRUE)
	{	
		for(i=1; i <= nsize-1; i++)			// do 13
		{
			k = i;
			tmp = eval[i];

			for(j = i+1; j <= nsize; j++)	// do 11
			{
				if(eval[j] <= tmp)
				{
					k = j;
					tmp = eval[j];
				}
			}								// 11 continue
			if(k != i)
			{
				eval[k] = eval[i];
				eval[i] = tmp;
				for(j = 1; j <= nsize; j++)	// do 12
				{
					tmp = evec[j][i];
					evec[j][i] = evec[j][k];
					evec[j][k] = tmp;
				}							// 12 continue
			}
		}									// 13 continue
	}
	else
	{
		for(i=1; i <= nsize-1; i++)			// do 23
		{
			k = i;
			tmp = eval[i];
			for(j = i+1; j <= nsize; j++)	// do 21
			{
				if(eval[j] >= tmp)
				{
					k = j;
					tmp = eval[j];
				}
			}								// 21 continue
			if(k != i)
			{
				eval[k] = eval[i];
				eval[i] = tmp;
				for(j = 1; j <= nsize; j++)	// do 22
				{
					tmp = evec[j][i];
					evec[j][i] = evec[j][k];
					evec[j][k] = tmp;
				}							// 22 continue
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////  

