#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "matutil.h"

#include "util.h"		//  Spinworks specific, for sw_error function

//  resolve possible MFC conflict
#ifdef ERROR
#undef ERROR
#endif
#define ERROR  -1

#ifdef FALSE
#undef FALSE
#endif
#define FALSE  0

//#define DUMP_LOG

//////////////////////////////////////////////////////////////////////////////////
//
//  Matrix utilities for SpinWorks
//  Mostly adapted from Numerical Recipes in C, with heavy modification
//	Most use unit based counting (FORTRAN style)
//
//////////////////////////////////////////////////////////////////////////////////

int *ivector(int nl, int nh)
{
int *v;

#ifdef DUMP_LOG
	FILE *ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "   ***  ivector %d  %d\n", nl, nh);
	fclose(ofp);
#endif

	v = (int *)malloc((unsigned)(nh-nl+1)*sizeof(int));

	if(v == NULL)
	{
		printf("cannot allocate %d  to  &d integer vector\n", nl, nh);
	}

	return v-nl;
}

//////////////////////////////////////////////////////////////////////////////////

double *vector(int nl, int nh)
{
	double *v;
	
#ifdef DUMP_LOG
	FILE *ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "   ***  vector  %d  %d\n", nl, nh);
	fclose(ofp);
#endif

	v=(double *)malloc((unsigned)(nh-nl+1)*sizeof(double));

	if(v == NULL)
	{
		printf("cannot allocate %d  to  &d double vector\n", nl, nh);
	}

	return v-nl;
}

//////////////////////////////////////////////////////////////////////////////////

void free_ivector(int *v, int nl, int nh)
{

#ifdef DUMP_LOG
	FILE *ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "   ***  free_ivector  %d  %d\n", nl, nh);
	fclose(ofp);
#endif

	if(v == NULL) 
	{
		printf("Warning: free_ivector(): attempt to free null reference");
		return;
	}
	free((char*)(v+nl));
}
    
/////////////////////////////////////////////////////////////////////////////////////////

void free_vector(double *v, int nl, int nh)
{

#ifdef DUMP_LOG
	FILE *ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "   ***  free_vector  %d  %d\n", nl, nh);
	fclose(ofp);
#endif

	if(v == NULL) 
	{
		printf("Warning: free_vector(): attempt to free null reference");
		return;
	}

	free((char*)(v+nl));
}
     
/////////////////////////////////////////////////////////////////////////////////////////

double **matrix(int nrl, int nrh, int ncl, int nch)
{
int i;
double **m;

#ifdef DUMP_LOG
	FILE *ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "   ***  matrix  %d  %d  %d  %d\n", nrl, nrh, ncl, nch);
	fclose(ofp);
#endif

	//  allocate pointer to rows
	m = (double **)malloc((unsigned)(nrh-nrl+1)*sizeof(double*));
	if(m == NULL) sw_error("matrix(): cannot allocate pointer to rows");
	m -= nrl;

	//  allocate rows and set pointers to them
	for(i=nrl; i<=nrh; i++)
    {
		m[i] = (double *)malloc((unsigned)(nch-ncl+1)*sizeof(double));
		if(m[i] == NULL)
		{
			sw_error("matrix(): cannot allocate row");
			break;
		}
		m[i] -= ncl;
    } 
	
	if(m == NULL)
	{
		printf("cannot allocate %d  by  &d double matrix\n", nrh - nrl, nch - ncl);
	}

	return m;
}

/////////////////////////////////////////////////////////////////////////////////////////

float **fmatrix(int nrl, int nrh, int ncl, int nch)

{
	int i;
	float **m;

#ifdef DUMP_LOG
	FILE *ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "   ***  matrix  %d  %d  %d  %d\n", nrl, nrh, ncl, nch);
	fclose(ofp);
#endif

	//  allocate pointer to rows
	m = (float **)malloc((unsigned)(nrh-nrl+1)*sizeof(float*));
	if(m == NULL) sw_error("matrix(): cannot allocate pointer to rows");
	m -= nrl;

	//  allocate rows and set pointers to them
	for(i=nrl; i<=nrh; i++)
    {
		m[i] = (float *)malloc((unsigned)(nch-ncl+1)*sizeof(float));
		if(m[i] == NULL)
		{
			sw_error("matrix(): cannot allocate row");
			break;
		}
		m[i] -= ncl;
    } 
	
	if(m == NULL)
	{
		printf("cannot allocate %d  by  &d float matrix\n", nrh - nrl, nch - ncl);
	}

	return m;
}

/////////////////////////////////////////////////////////////////////////////////////////

int **imatrix(int nrl, int nrh, int ncl, int nch)
//    int nrl, nrh, ncl, nch;

{
	int i;
	int **m;

#ifdef DUMP_LOG
	FILE *ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "   ***  imatrix  %d  %d  %d  %d\n", nrl, nrh, ncl, nch);
	fclose(ofp);
#endif

	m = (int **)malloc((unsigned)(nrh-nrl+1)*sizeof(int*));
	if(m == NULL) sw_error("imatrix(): cannot allocate pointer to rows");
	m -= nrl;

	for(i=nrl; i<=nrh; i++)
    {
		m[i] = (int *)malloc((unsigned)(nch-ncl+1)*sizeof(int));
		if(m[i] == NULL)
		{
			sw_error("imatrix(): cannot allocate row");
			break;
		}
		m[i] -= ncl;
    }
    
	if(m == NULL)
	{
		printf("cannot allocate %d  by  &d integer matrix\n", nrh - nrl, nch - ncl);
	}

	return m;
}

/////////////////////////////////////////////////////////////////////////////////////////

void  free_matrix(double **m, int nrl, int nrh, int ncl, int nch)
    
{
	int i;

#ifdef DUMP_LOG
	FILE *ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "   ***  free_matrix  %d  %d  %d  %d\n", nrl, nrh, ncl, nch);
	fclose(ofp);
#endif

	if(m == NULL)
	{
		printf("Warning: free_matrix(): attempt to free null reference");
		return;
	}

	for(i=nrh; i>=nrl; i--)
		free((char*)(m[i]+ncl));
	free((char*)(m+nrl));
}

/////////////////////////////////////////////////////////////////////////////////////////
void  free_imatrix(int **m, int nrl, int nrh, int ncl, int nch)
    
{
	int i;

#ifdef DUMP_LOG
	FILE *ofp = fopen("matdump.txt", "a");
	fprintf(ofp, "   ***   free_matrix  %d  %d  %d  %d\n", nrl, nrh, ncl, nch);
	fclose(ofp);
#endif

	if(m == NULL)
	{
		printf("Warning: free_imatrix(): attempt to free null reference");
		return;
	}

	for(i=nrh; i>=nrl; i--)
		free((char*)(m[i]+ncl));
	free((char*)(m+nrl));
}

/////////////////////////////////////////////////////////////////////////////////////////


void matclear(double **m, int nrl, int nrh, int ncl, int nch, 
			  double value)
{
	int i, j;
	for(i=nrl; i<= nrh; i++)
	{
		for(j=ncl; j <= nch; j++)
		{
			m[i][j] = value;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////

void imatclear(int **m, int nrl, int nrh, int ncl, int nch, 
			  int value)
{
	int i, j;
	for(i=nrl; i<= nrh; i++)
	{
		for(j=ncl; j <= nch; j++)
		{
			m[i][j] = value;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////

void dclear(int arrsiz, double *array, double value)
{
	int i;
	for(i=1; i<=arrsiz; i++)
	{
		array[i] = value;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////

void iclear(int arrsiz, int *array, int value)
{
	int i;
	for(i=1; i<=arrsiz; i++)
	{
		array[i] = value;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////

void matdump(double **m, int nrl, int nrh, int ncl, int nch, char *title)
{
//	Dump a double precision matrix to a file
//  Used only for debugging
	int i, j;
	FILE *ofp = fopen("matdump.txt", "a");

	if(ofp == NULL)
	{
		printf("Warning: matdump(): cannot open matdump.txt");
		return;
	}

	fprintf(ofp, "\n\n*****************  matrix dump *****************\n");
	fprintf(ofp, "***  of %s\n", title);
	for(i=nrl; i<= nrh; i++)
	{
		for(j=ncl; j <= nch; j++)
		{
			fprintf(ofp, "  %10.3f", m[i][j]);
		}
		fprintf(ofp, "\n");
	}
	fprintf(ofp, "\n************************************************\n");
	fclose(ofp);
}

/////////////////////////////////////////////////////////////////////////////////////////

void imatdump(int **m, int nrl, int nrh, int ncl, int nch)
{
//	Dump an integer matrix to a file
//  Used only for debugging
	int i, j;
	FILE *ofp = fopen("matdump.txt", "a");

	if(ofp == NULL)
	{
		printf("Warning: imatdump(): cannot open matdump.txt");
		return;
	}

	fprintf(ofp, "\n*****************  matrix dump *****************\n");
	for(i=nrl; i<= nrh; i++)
	{
		for(j=ncl; j <= nch; j++)
		{
			fprintf(ofp, "  %5d", m[i][j]);
		}
		fprintf(ofp, "\n");
	}
	fprintf(ofp, "\n***********************************************\n");
	fclose(ofp);
}

/////////////////////////////////////////////////////////////////////////////////////////


int binint(int maxsize, int cursize, int *array, int key)
{
//	This function does a binary search of an array of integers for the
//	key specified by "key".  It returns the subscript if found, zero
//	otherwise.

	int ret = 0;
	int top = 0;
	int bottom = cursize + 1;
	int midpoint;

	while( bottom - top >= 2)
	{
		midpoint = (top + bottom)/2;
		if(key < array[midpoint])
		{
			bottom = midpoint;
		}
		else 
		{
			if(key > array[midpoint])
			top = midpoint;
			else
			{
				ret = midpoint;
				bottom = top;
			}
		}
	}	// while

	return ret;
}

/////////////////////////////////////////////////////////////////////////////////////////


double fmax(double a, double b)
{
	if(a >= b) 
		return a;
	else
		return b;
}

/////////////////////////////////////////////////////////////////////////////////////////

double fmin(double a, double b)
{
	if(a <= b) 
		return a;
	else
		return b;
}


/////////////////////////////////////////////////////////////////////////////////////////

int imax(int a, int b)
{
	if(a >= b) 
		return a;
	else
		return b;
}

/////////////////////////////////////////////////////////////////////////////////////////

int imin(int a, int b)
{
	if(a <= b) 
		return a;
	else
		return b;
}

/////////////////////////////////////////////////////////////////////////////////////////

int ivect_min(int * array, int size)
{
	int i, min_val = 100000000;

	for(i = 0; i < size; i++)
		if(array[i] < min_val) min_val = array[i];

	return min_val;
}

/////////////////////////////////////////////////////////////////////////////////////////

int pow2(int n)
{
	int result = 1;
	for(int i = 1; i <= n; i++)
		result = result * 2;
	return result;
}
/////////////////////////////////////////////////////////////////////////////////////////

double minmax(double minval, double curval, double maxval)
{
//	This routine checks to see that CURVAL is in the range MINVAL
//	to MAXVAL, returning one of the range limits if it is not.

	double temp;

	temp = curval;

	if (minval < maxval)		// in proper order
	{
		if(curval < minval) temp = minval;
		if(curval > maxval) temp = maxval;
	}
	else						// in reverse order
	{
		if(curval < maxval) temp = maxval;
		if(curval > minval) temp = minval;
	}

	return temp;
}

/////////////////////////////////////////////////////////////////////////////////////////

void symto(int n, int *pairg, double **asmmat, double **symmat)
{
//     This routine transforms the asymmetric matrix ASYMMAT into
//     the symmetrized matrix SASMMAT.

	int i, j;

	for(i=1; i<=n; i++)
	{
		symmat[i][i] = asmmat[i][i];
		for(j=i+1; j<=n; j++)
		{
			if(pairg[i] == 2 && pairg[j] == 2)
			{
				symmat[i][j] = 0.5 * (asmmat[i][j] + asmmat[j][i]);
				symmat[j][i] = 0.5 * (asmmat[i][j] - asmmat[j][i]);
			}
			else
			{
				symmat[i][j] = asmmat[i][j];
				symmat[j][i] = asmmat[j][i];
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////

void symfrm(int n, int *pairg, double **symmat, double **asmmat)
{
//     This routine transforms the symmetrized matrix SYMMAT into
//     the 'asymmetric' matrix ASMMAT.

	int i, j;

	for(i = 1; i <= n; i++)
	{
		asmmat[i][i] = symmat[i][i];
		for(j = i+1; j <= n; j++)
		{
			if(pairg[i] == 2 && pairg[j] == 2)
			{
				asmmat[i][j] = 1.0 * (symmat[i][j] + symmat[j][i]);
				asmmat[j][i] = 1.0 * (symmat[i][j] - symmat[j][i]);
			}
			else
			{
				asmmat[i][j] = symmat[i][j];
				asmmat[j][i] = symmat[j][i];
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////

int ludcmp(int n, double **mat, int *index, double *scale, double *xchng)
{
//   Given an N by N matrix MAT, with physical dimension NP, this
//   routine replaces it by the LU decomposition of a rowwise
//   permutation of itself. INDX is an output vector which records
//   the row permutation effected by the partial pivoting. XCHNG is
//   output as +/- 1.0 depending on whether the number of row
//   interchanges was even or odd, respectively. ERROR is output
//   as .true. if the matrix is singular. SCALE is a scratch vector
//   that stores the implicit scaling of each row.

//   This routine is used in combination with LUDBAK to solve
//   linear equations, or invert a matrix.

	int		error = FALSE;
	int		i, j, k, ipiv;
	double  tiny = 1.0e-34, aamax, tmp, sum;

	*xchng = 1.0;

	for(i = 1; i <= n; i++)
	{
		aamax = 0.0;
		for(j = 1; j <= n; j++)
		{
			tmp = fabs(mat[i][j]);
			if(tmp > aamax) aamax = tmp;
		}
		if(aamax == 0.0)
		{
			return ERROR;
		}
		scale[i] = 1.0 / aamax;
	}

	for(j = 1; j <= n; j++)
	{
		if(j > 1)
		{
			for(i = 1; i <= j-1; i++)
			{
				sum = mat[i][j];
				if(i > 1)
				{
					for(k = 1; k <= i-1; k++)
						sum = sum - mat[i][k]*mat[k][j];
					mat[i][j] = sum;
				}
			}
		}
		
		aamax = 0.0;
		ipiv = j;
		for(i = j; i <= n; i++)
		{
			sum = mat[i][j];
			if(j > 1)
			{
				for(k = 1; k <= j-1; k++)
					sum = sum - mat[i][k] * mat[k][j];
				mat[i][j] = sum;
			}
			tmp = scale[i]*fabs(sum);
			if(tmp >= aamax)
			{
				ipiv = i;
				aamax = tmp;
			}
		}

		if(j != ipiv)
		{
			for(k = 1; k <= n; k++)
			{
				tmp = mat[ipiv][k];
				mat[ipiv][k] = mat[j][k];
				mat[j][k] = tmp;
			}

			*xchng = -(*xchng);
			scale[ipiv] = scale[j];
		}
		index[j] = ipiv;

//  Divide by the pivot element. If the pivot element is zero
//  the matrix is singular (at least to the precision of the
//  algorithm). For some applications on singular matrices, it
//  is desirable to substitutea finite small value for zero.

		if(j != n)
		{
			if(mat[j][j] == 0.0) mat[j][j] = tiny;
			tmp = 1.0 / mat[j][j];
			for(i = j+1; i <= n; i++)
				mat[i][j] = mat[i][j] * tmp;
		}
	}
	if(mat[n][n] == 0.0) mat[n][n] = tiny;
	return error;
}

/////////////////////////////////////////////////////////////////////////////////////////

void ludbak(int n, double **mat, int *index, double *rhs)
{
//   This routine solves the set of N linear equations MAT.X = RHS
//   A is input not as the matrix, but as its LU decomposition,
//   determined by the routine LUDCMP. INDX is the permutation
//   vector returned by LUDCMP. RHS is the right-handside vector,
//   and returns with the solution vector. NP, N, MAT, and INDX
//   are not modified and can be left in place for successive
//   calls with different RHS's. This routine takes into account
//   the possibility that RHS will begin with many zero elements,
//   so it is efficient for use in matrix inversion.

	int i, ii, j, ll;
	double sum;

	ii = 0;
	for(i = 1; i <= n; i++)
	{
		ll = index[i];
		sum = rhs[ll];
		rhs[ll] = rhs[i];

		if(ii != 0)
		{
			for(j = ii; j <= i-1; j++)
				sum = sum - mat[i][j] * rhs[j];
		}
		else if(sum != 0.0)
		{
			ii = i;
		}
		rhs[i] = sum;
	}

	// now back substitute
	for(i = n; i >= 1; i--)
	{
		sum = rhs[i];
		if(i , n)
		{
			for(j = i+1; j <= n; j++)
				sum = sum - mat[i][j] * rhs[j];
		}
		rhs[i] = sum / mat[i][i];
	}
}

/////////////////////////////////////////////////////////////////////////////////////////

void matident(int size, double **mat)
{
//	clears and sets matrix mat to the identity matrix
	int i;

	matclear(mat, 1, size, 1, size, 0.0);

	for(i = 1; i <= size; i++)
		mat[i][i] = 1.0;
}

/////////////////////////////////////////////////////////////////////////////////////////

int invert(int size, double **mat, double **inv)
{
//  returns the inverse of mat in inv
//  mat is not altered, returns ERROR if matrix is singular
	int error = FALSE;
	int i, j, *pindx;
	double **work_mat;							// temporary working matrix
	double *scale, *col, xchng;

	work_mat = matrix(1, size, 1, size);		// allocate space
	scale = vector(1, size);
	col = vector(1, size);
	pindx = ivector(1, size);

	matident(size, inv);						// set inv to identity matrix

	for(i=1; i <= size; i++)					// copy matrix
		for(j = 1; j <= size; j++)
			work_mat[i][j] = mat[i][j];

	error = ludcmp(size, work_mat, pindx, scale, &xchng);

	if(error == FALSE)
	{
		for(j = 1; j <= size; j++)
		{
			for(i = 1; i <= size; i++)
			{
				col[i] = 0.0;
			}
			col[j] = 1.0;
			ludbak(size, work_mat, pindx, col);
			for(i = 1; i <= size; i++)
			{
				inv[i][j] = col[i];
			}
		}
	}

	free_matrix(work_mat, 1, size, 1, size);	// free-up space
	free_vector(scale, 1, size);
	free_vector(col, 1, size);
	free_ivector(pindx, 1, size);
	return error;
}

#ifdef DUMP_LOG
#undef DUMP_LOG
#endif

