// Functions used for transition input and output

#include <stdio.h>
#include "nummrit.h"
#include "maxvalue.h"
#include "matutil.h"
#include "sort_util.h"
#include "trnout.h"
#include "adjust.h"

// /grpcom/
extern double	*cshift, **cup, **symcup,
				**dip, **symdip;
extern int		*grpiso, *nucg, *pairg,
				*grpspn, fmaxg[MAXGRP+1][MAXGRP+1], *agpb;
extern short	anydip, symspn;
extern char		grpnam[MXGNAM + 1][MAXGRP + 1];


// /spncom/
extern int		ngrp, niso, nvar;

// /isocom/
extern double	*tiiso, *tliso,
				*fliso, *fuiso;
extern int		*isospn, *ntiso,
				*misopa, *misopb, km2p;
extern char		**isonam;
extern char		*prtiso;

// /flgcom/
extern short	autasn, autdel, optmiz, prtptl,
				prtkey, prtdif, prtvcm, pracnd;

// /shpcom/
extern double	shpfac, ampscl;

// /asncom/
extern double	*asnfrq, *asnwgt, *asndif, *asnclc, asnint;
extern int		asncnt, *asnlin;

// /optcom/
extern double	rmscon, rmsdiv, rmslim, damp, asnlim,
				ignlim, itrfac, linwid, asnrms;
extern int		itrlim, itrmin, asnmod;
extern short	anyasn, wgtrms;

//extern CString sim_out;

/////////////////////////////////////////////////////////////////////////////////

int tran_out(const char *sim_out, const char *trans_out)
{
	double *calc_freq, *calc_amp;
	int *calc_line, *calc_key;
	int calc_count;
	double total_amplitude;

	calc_freq = vector(1, MAXCLC);			// allocate space
	calc_amp = vector(1, MAXCLC);
	calc_line = ivector(1, MAXCLC);
	calc_key = ivector(1, MAXCLC);

	dclear(MAXCLC, calc_freq, 0.0);			// initialize
	dclear(MAXCLC, calc_amp, 0.0);
	iclear(MAXCLC, calc_line, 0);
	iclear(MAXCLC, calc_key, 0);

	int error = FALSE;

	for(int i = 1; i <= niso; i++)
	{
		error = tran_get(i, MAXCLC, &calc_count, &total_amplitude,
			             calc_freq, calc_amp, calc_line, calc_key);

		shell_sort_trans(MAXCLC, calc_count, calc_freq, calc_amp,
				calc_line, calc_key);

		print_tran(sim_out, i, calc_count, calc_freq, calc_amp,
				 calc_line,  calc_key, total_amplitude);

		plot_tran(sim_out, trans_out, i, MAXCLC, calc_count, calc_freq, 
			  calc_amp, calc_line);
	}

//	output optimized coupling and shift info at end of transition file
	FILE *outfile = fopen(trans_out, "a");
	fprintf(outfile, "$$end of transitions\n");

	int i, j;
	for(i = 1; i <= ngrp; i++)
	{
		fprintf(outfile, "v %d = %10.5f\n", i, cshift[i]);
	}

	for(i = 1; i <=ngrp; i++)
	{
		for(int j = 1; j <=ngrp; j++)
		{
			if(cup[i][j] != 0.0) fprintf(outfile, "j %d %d = %9.6f\n", i, j, cup[i][j]);
			if(dip[i][j] != 0.0) fprintf(outfile, "d %d %d = %9.6f\n", i, j, dip[i][j]);
		}
	}
	fclose(outfile);

	if(optmiz)group_rms(sim_out);

	free_vector(calc_freq, 1, MAXCLC);			// free space
	free_vector(calc_amp, 1, MAXCLC);
	free_ivector(calc_line, 1, MAXCLC);
	free_ivector(calc_key, 1, MAXCLC);

	return error;
}

/////////////////////////////////////////////////////////////////////////////////

int tran_get(int iso, int maxtran, int *p_calc_count, double *p_tran_amp, 
			  double *calc_freq,  double *calc_amp,
			  int *calc_line, int *calc_key)

{

	int		km2, km32m1, km32, isospc,
			line_no, mpacka, m4a, levela, m4b, levelb;

	double	freq, ampl, illim, fllim, fulim;
	char	dummy[20];
	FILE	*isoscr;
	int i;
	double total_amplitude = 0.0;

//	open transition file
	sprintf(dummy, "isoscr.%d", iso);
	isoscr = fopen(dummy, "rb");
	if(isoscr == NULL)
	{
		*p_tran_amp = 0.0;
		*p_calc_count = 0;
		return ERROR;
	}

	fllim = fliso[iso];
	fulim = fuiso[iso];
	illim = tliso[iso];

	i = 0;
	while(feof(isoscr) == 0 && ferror(isoscr) == 0  && i <= maxtran)
	{
		fread(&km2, sizeof(int), 1, isoscr);
		fread(&km32m1, sizeof(int), 1, isoscr);
		fread(&km32, sizeof(int), 1, isoscr);
		fread(&isospc, sizeof(int), 1, isoscr);
		fread(&line_no, sizeof(int), 1, isoscr);
		fread(&freq, sizeof(double), 1, isoscr);
		fread(&ampl, sizeof(double), 1, isoscr);
		fread(&mpacka, sizeof(int), 1, isoscr);
		fread(&m4a, sizeof(int), 1, isoscr);
		fread(&levela, sizeof(int), 1, isoscr);
		fread(&m4b, sizeof(int), 1, isoscr);
		fread(&levelb, sizeof(int), 1, isoscr);

		if((ampl >= illim) && (freq >= fllim) && (freq <= fulim)
			&& feof(isoscr) == 0 && ferror(isoscr) == 0)
		{
			i++;
			total_amplitude = total_amplitude + ampl;
			calc_freq[i] = freq;
			calc_amp[i] = ampl;
			calc_line[i] = line_no;
			calc_key[i] = mpacka + km2*km2p;
		}
	}

	*p_tran_amp = total_amplitude;
	*p_calc_count = i;
	fclose(isoscr);
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////////

int print_tran(const char *sim_out, int iso, int calc_count, double *calc_freq, double *calc_amp,
				int *calc_line, int *calc_key, double total_amplitude)
{
	FILE *outfile;
	int ixasn;

	outfile = fopen(sim_out, "a");

	if(outfile == NULL)
	{
		return ERROR;
	}

	fprintf(outfile, "\n Transitions for isotopic species %d\n\n", iso);

	for(int i = 1; i <= calc_count; i++)
	{
		ixasn = 0; 
		if(optmiz == TRUE)
			ixasn = binint(MAXASN, asncnt, asnlin, calc_line[i]);

		if(optmiz == TRUE)
		{
			if(ixasn != 0)
				fprintf(outfile, " %6d  %9.3f  %9.3f %9.3f %9.5 %9.5f\n",
			    calc_line[i], asnfrq[ixasn], calc_freq[i], asndif[ixasn],
				calc_amp[i], asnwgt[ixasn]);
			else
				fprintf(outfile, " %6d             %9.3f            %9.5f \n",
					calc_line[i], calc_freq[i], calc_amp[i]);
		}
		else
			fprintf(outfile, " %6d  %9.3f %9.3f \n",
			    calc_line[i],  calc_freq[i], calc_amp[i]);

	}
	fclose(outfile);
	return FALSE;
}
/////////////////////////////////////////////////////////////////////////////////////////

int plot_tran(const char *sim_out, const char *trans_out, int iso, int maxtrn, 
			  int calc_count, double *calc_freq, double *calc_amp, int *calc_line)
{
	FILE *outfile;
	int i, j;

	if(iso == 1)
		outfile = fopen(trans_out, "w");
	else
		outfile = fopen(trans_out, "a");

	for(i=1; i <= calc_count; i++)
	{
		fprintf(outfile, "%12.4f     %10.5f      %10.5f     %11d\n",
			    calc_freq[i], ampscl*calc_amp[i], shpfac, calc_line[i]);
	}

	fclose(outfile);

	FILE *ofp = fopen(sim_out, "a");
	fprintf(ofp, "\n***  %d transitions for species %d written to plot file\n", i-1, iso);
	fclose(ofp);

	return FALSE;
}


